# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__: typing.Final[typing.List[str]] = [
    "Option",
    "BaseSlashCommand",
    "WithAsyncCallback",
    "WithGetCommand",
    "WithCreationMethods",
    "WithGetOptions",
    "WithAsOption",
    "SlashCommand",
    "SlashCommandGroup",
    "SlashSubGroup",
    "SlashSubCommand",
]

import abc
import collections.abc
import dataclasses
import functools
import logging
import typing
import warnings

import hikari

if typing.TYPE_CHECKING:
    from lightbulb import command_handler
    from lightbulb.slash_commands import context as context_

_LOGGER = logging.getLogger("lightbulb")

OPTION_TYPE_MAPPING = {
    str: hikari.OptionType.STRING,
    int: hikari.OptionType.INTEGER,
    bool: hikari.OptionType.BOOLEAN,
    hikari.User: hikari.OptionType.USER,
    hikari.TextableChannel: hikari.OptionType.CHANNEL,
    hikari.Role: hikari.OptionType.ROLE,
    hikari.Snowflake: hikari.OptionType.MENTIONABLE,
    float: hikari.OptionType.FLOAT,
}


def _serialise_option(opt: hikari.CommandOption):
    data = {
        "name": opt.name.lower(),
        "description": opt.description,
        "type": opt.type,
        "required": opt.is_required,
    }

    if opt.options is not None:
        data["options"] = list(
            sorted((_serialise_option(sub_option) for sub_option in opt.options), key=lambda o: o["name"])
        )
    elif opt.choices is not None:
        data["choices"] = list(
            sorted(({"name": c.name.lower(), "value": c.value} for c in opt.choices), key=lambda c: c["name"])
        )
    else:
        data["options"] = []

    return data


def _serialise_command(cmd: typing.Union[hikari.Command, typing.Union[SlashCommand, SlashCommandGroup]]):
    opts = cmd.options if isinstance(cmd, hikari.Command) else cmd.get_options()
    return {
        "name": cmd.name.lower(),
        "description": cmd.description,
        "options": list(sorted((_serialise_option(opt) for opt in opts), key=lambda o: o["name"]))
        if opts is not None
        else [],
    }


def _resolve_subcommand(
    context: context_.SlashCommandContext,
    subcommands: typing.MutableMapping[str, typing.Union[SlashSubCommand, SlashSubGroup]],
) -> typing.Optional[str]:
    option_name = None
    for option in context.options:
        if option in subcommands:
            option_name = option
            break
    return option_name


@dataclasses.dataclass
class Option:
    """
    Dataclass representing a command option.

    Examples:
        Usage in a slash command:

        .. code-block:: python

            class Echo(SlashCommand):
                description = "Repeats the input"
                # Options
                text: str = Option("Text to repeat")

                async def callback(self, context):
                    await context.respond(context.option_values.text)

    """

    description: str
    """The description of the option."""
    name: str = None
    """The name of the option. If ``None`` then this will be the name of the attribute."""
    required: bool = None
    """Whether or not the option is required. If ``None`` then it will be inferred from the attribute's typehint."""


def _get_type_and_required_from_option(hint, opt: Option) -> typing.Tuple[hikari.OptionType, bool]:
    type_ = OPTION_TYPE_MAPPING[typing.get_args(hint)[0] if typing.get_args(hint) else hint]
    required = opt.required if opt.required is not None else (type(None) in typing.get_args(hint))
    return type_, required


def _get_options_for_command_instance(
    cmd: typing.Union[SlashCommand, SlashSubCommand]
) -> typing.Sequence[hikari.CommandOption]:
    if (
        hasattr(cmd, "options")
        and isinstance(cmd.options, collections.abc.Sequence)
        and not isinstance(cmd.options, str)
    ):
        warnings.warn(
            "Definition of command options in the 'options' attribute is deprecated and "
            "scheduled for removal in version 1.4. You should define options using class variables instead.",
            DeprecationWarning,
        )
        return cmd.options

    all_attrs = [[attr_name, getattr(cmd, attr_name)] for attr_name in dir(cmd)]
    opts = filter(lambda opt: type(opt[1]) is Option, all_attrs)
    hints = typing.get_type_hints(cmd)

    hk_options = []
    for attr_name, option in opts:
        type_, required = _get_type_and_required_from_option(hints.get(attr_name), option)
        hk_options.append(
            hikari.CommandOption(
                name=option.name or attr_name, description=option.description, type=type_, is_required=required
            )
        )
    return hk_options


class BaseSlashCommand(abc.ABC):
    """
    Abstract base class for slash command-like classes.

    Args:
        bot (:obj:`~lightbulb.command_handler.Bot`): The bot instance the command will be added to.
    """

    __slots__: typing.Sequence[str] = ("bot", "_instances")

    def __init__(self, bot: command_handler.Bot) -> None:
        self.bot = bot
        """The bot instance that the slash command is registered to."""
        self._instances: typing.MutableMapping[hikari.Snowflakeish, hikari.Command] = {}

    @property
    def name(self) -> str:
        """
        The name of the slash command. Defaults to the class name converted to lowercase. Can be a maximum
        of 32 characters long.

        Returns:
            :obj:`str`: Slash command name.
        """
        return self.__class__.__name__.lower()

    @property
    @abc.abstractmethod
    def description(self) -> str:
        """
        The description of the slash command. Can be a maximum of 100 characters long.

        Returns:
            :obj:`str`: Slash command description.
        """
        ...


class WithAsyncCallback(abc.ABC):
    @abc.abstractmethod
    async def callback(self, context: context_.SlashCommandContext) -> None:
        """
        The slash command callback method. This method will be called whenever the slash command is invoked.

        If the slash command being invoked is a subcommand then the
        :obj:`~lightbulb.slash_commands.SlashCommandContext.options`  attribute will have been replaced by the options
        that the subcommand was invoked with, instead of those that the command as a whole was invoked with  (they can
        still be accessed through :obj:`~lightbulb.slash_commands.SlashCommandContext.interaction.options` if necessary).

        Args:
            context (:obj:`~lightbulb.slash_commands.SlashCommandContext`): The context that the slash command
                was invoked under.

        Returns:
            ``None``
        """
        ...


class WithAsOption(abc.ABC):
    @abc.abstractmethod
    def as_option(self) -> hikari.CommandOption:
        """
        Creates and returns the appropriate :obj:`hikari.CommandOption` representation for this
        subcommand class.

        Returns:
            :obj:`hikari.CommandOption`: The ``CommandOption`` version of the subcommand class.
        """
        ...


class WithGetOptions(abc.ABC):
    @property
    def enabled_guilds(self) -> typing.Optional[typing.Union[hikari.Snowflakeish, hikari.SnowflakeishSequence]]:
        """
        The guilds that the slash command is enabled in. If ``None`` or an empty sequence, the command will be
        added as a global command. Defaults to an empty list, therefore making the command global unless otherwise
        specified.

        Returns:
            Optional[Union[:obj:`hikari.Snowflakeish`, :obj:`hikari.SnowflakeishSequence`]]: Guilds that the command
                is enabled in, or ``None`` or empty sequence if the command is global.
        """
        return []

    @abc.abstractmethod
    def get_options(self) -> typing.Sequence[hikari.CommandOption]:
        """
        Get the options for the command.

        Returns:
            Sequence[:obj:`hikari.CommandOption`]: Options for the command.
        """
        ...


class WithCreationMethods(abc.ABC):
    @abc.abstractmethod
    async def create(
        self,
        app: hikari.SnowflakeishOr[hikari.PartialApplication],
        guild_id: typing.Optional[hikari.Snowflakeish] = None,
    ) -> hikari.Command:
        """
        Creates the command for a specific guild, or globally if no guild ID is given.

        Args:
            app (hikari.SnowflakeishOr[hikari.PartialApplication]): The application to add the command to.
            guild_id (Optional[:obj:`int`]): The ID of the guild to create the command for.

        Returns:
            :obj:`hikari.Command`: The command object created.
        """
        ...

    async def delete(
        self,
        app: hikari.SnowflakeishOr[hikari.PartialApplication],
        guild_id: typing.Optional[hikari.Snowflakeish] = None,
    ) -> None:
        """
        Deletes the command for a specific guild, or globally if no guild ID is given.

        Args:
            app (hikari.SnowflakeishOr[hikari.PartialApplication]): The application to delete the command from.
            guild_id (Optional[hikari.Snowflakeish]): The ID of the guild to delete the command for.

        Returns:
            ``None``
        """
        await self.bot.rest.delete_application_command(
            app, self._instances[guild_id], **({"guild": guild_id} if guild_id is not None else {})
        )
        self._instances.pop(guild_id)

    async def auto_create(self, app: hikari.SnowflakeishOr[hikari.PartialApplication]) -> None:
        """
        Creates the command in the appropriate scopes (guilds, global) automatically.

        Args:
            app (hikari.SnowflakeishOr[hikari.PartialApplication]): The application to add the command to.

        Returns:
            ``None``
        """
        if not self.enabled_guilds:
            await self.create(app)
        else:
            for guild_id in self.enabled_guilds:
                await self.create(app, guild_id)

    async def auto_delete(self, app: hikari.SnowflakeishOr[hikari.PartialApplication]) -> None:
        """
        Deletes the command from the appropriate scopes (guilds, global) automatically.

        Args:
            app (hikari.SnowflakeishOr[hikari.PartialApplication]): The application to remove the command from.

        Returns:
            ``None``
        """
        if not self.enabled_guilds:
            await self.delete(app)
        else:
            for guild_id in self.enabled_guilds:
                await self.delete(app, guild_id)


class WithGetCommand(abc.ABC):
    def get_command(self, guild_id: typing.Optional[hikari.Snowflakeish] = None) -> typing.Optional[hikari.Command]:
        """
        Gets the :obj:`hikari.Command` instance of this command class for a given ``guild_id``, or the global
        instance if no ``guild_id`` is provided. Returns ``None`` if no instance is found for the given ``guild_id``.

        Args:
            guild_id (Optional[:obj:`hikari.Snowflakeish`): The guild ID to get the :obj:`hikari.Command` instance
                for.

        Returns:
            Optional[:obj:`hikari.Command`]: Command instance for that guild, or ``None`` if no instance exists.
        """
        return self._instances.get(guild_id)


class SlashCommand(BaseSlashCommand, WithGetOptions, WithAsyncCallback, WithCreationMethods, WithGetCommand, abc.ABC):
    """
    Abstract base class for top level slash commands. All slash commands that are not groups
    should inherit from this class.

    All abstract methods **must** be implemented by your custom slash command class. A list of the abstract
    methods and properties you are required to implement for this class can be seen below:

    - :obj:`~lightbulb.slash_commands.BaseSlashCommand.description` (class variable)
    - :obj:`~lightbulb.slash_commands.WithAsyncCallback.callback` (instance method)
    """

    __slots__ = ()

    async def __call__(self, *args, **kwargs):
        return await self.callback(*args, **kwargs)

    @functools.lru_cache
    def get_options(self) -> typing.Sequence[hikari.CommandOption]:
        hk_options = _get_options_for_command_instance(self)
        return hk_options[::-1]

    async def create(
        self,
        app: hikari.SnowflakeishOr[hikari.PartialApplication],
        guild_id: typing.Optional[hikari.Snowflakeish] = None,
    ) -> hikari.Command:
        created_command = await self.bot.rest.create_application_command(
            app,
            self.name,
            self.description,
            options=self.get_options(),
            **({"guild": guild_id} if guild_id is not None else {}),
        )
        self._instances[guild_id] = created_command
        return created_command


class SlashCommandGroup(BaseSlashCommand, WithGetOptions, WithCreationMethods, WithGetCommand, abc.ABC):
    """
    Abstract base class for slash command groups. All slash command groups should inherit from this class.

    All abstract methods **must** be implemented by your custom slash command group class. A list of the abstract
    methods and properties you are required to implement for this class can be seen below:

    - :obj:`~lightbulb.slash_commands.BaseSlashCommand.description` (class variable)
    """

    __slots__: typing.Sequence[str] = ("_subcommands",)

    _subcommand_list: typing.List[typing.Union[typing.Type[SlashSubCommand], typing.Type[SlashSubGroup]]] = []

    def __init__(self, bot: command_handler.Bot):
        super().__init__(bot)
        self._subcommands: typing.MutableMapping[str, SlashSubCommand] = {}
        for cmd_class in self._subcommand_list:
            cmd = cmd_class(bot)
            self._subcommands[cmd.name] = cmd

    @functools.lru_cache
    def get_options(self) -> typing.Sequence[hikari.CommandOption]:
        return [cmd.as_option() for cmd in self._subcommands.values()]

    async def __call__(self, context: context_.SlashCommandContext) -> None:
        option_name = _resolve_subcommand(context, self._subcommands)
        if option_name is None:
            return

        _LOGGER.debug("invoking slash subcommand %r", option_name)
        # Replace the context options with the options for the subcommand, the old options
        # can still be accessed through context.interaction.options
        new_options = context.options[option_name].options
        context.options = {option.name: option for option in new_options} if new_options is not None else {}
        return await self._subcommands[option_name](context)

    @classmethod
    def subcommand(cls) -> typing.Callable[[typing.Type[SlashSubCommand]], typing.Type[SlashSubCommand]]:
        """
        Decorator which registers a subcommand to this slash command group.
        """

        def decorate(subcommand_class: typing.Type[SlashSubCommand]) -> typing.Type[SlashSubCommand]:
            cls._subcommand_list.append(subcommand_class)
            return subcommand_class

        return decorate

    @classmethod
    def subgroup(cls) -> typing.Callable[[typing.Type[SlashSubGroup]], typing.Type[SlashSubGroup]]:
        """
        Decorator which registers a subgroup to this slash command group.
        """

        def decorate(subgroup_class: typing.Type[SlashSubGroup]) -> typing.Type[SlashSubGroup]:
            cls._subcommand_list.append(subgroup_class)
            return subgroup_class

        return decorate

    async def create(
        self,
        app: hikari.SnowflakeishOr[hikari.PartialApplication],
        guild_id: typing.Optional[hikari.Snowflakeish] = None,
    ) -> hikari.Command:
        created_command = await self.bot.rest.create_application_command(
            app,
            self.name,
            self.description,
            options=self.get_options(),
            **({"guild": guild_id} if guild_id is not None else {}),
        )
        self._instances[guild_id] = created_command
        return created_command


class SlashSubGroup(BaseSlashCommand, WithAsOption, abc.ABC):
    """
    Abstract base class for slash subgroups. All slash subgroups should inherit from this class.

    All abstract methods **must** be implemented by your custom slash subgroup class. A list of the abstract
    methods and properties you are required to implement for this class can be seen below:

    - :obj:`~lightbulb.slash_commands.BaseSlashCommand.description` (class variable)
    """

    __slots__: typing.Sequence[str] = ("_subcommands",)

    _subcommand_list: typing.List[typing.Type[SlashSubCommand]] = []

    def __init__(self, bot: command_handler.Bot):
        super().__init__(bot)
        self._subcommands: typing.MutableMapping[str, SlashSubCommand] = {}
        for cmd_class in self._subcommand_list:
            cmd = cmd_class(bot)
            self._subcommands[cmd.name] = cmd

    async def __call__(self, context: context_.SlashCommandContext) -> None:
        option_name = _resolve_subcommand(context, self._subcommands)
        if option_name is None:
            return

        _LOGGER.debug("invoking slash subcommand %r", option_name)
        # Replace the context options with the options for the subcommand, the old options
        # can still be accessed through context.interaction.options
        new_options = context.options[option_name].options
        context.options = {option.name: option for option in new_options} if new_options is not None else {}
        return await self._subcommands[option_name](context)

    @classmethod
    def subcommand(cls) -> typing.Callable[[typing.Type[SlashSubCommand]], typing.Type[SlashSubCommand]]:
        """
        Decorator which registers a subcommand to this slash command group.
        """

        def decorate(subcommand_class: typing.Type[SlashSubCommand]) -> typing.Type[SlashSubCommand]:
            cls._subcommand_list.append(subcommand_class)
            return subcommand_class

        return decorate

    @functools.lru_cache
    def as_option(self) -> hikari.CommandOption:
        return hikari.CommandOption(
            name=self.name,
            description=self.description,
            type=hikari.OptionType.SUB_COMMAND_GROUP,
            options=[cmd.as_option() for cmd in self._subcommands.values()],
            is_required=False,
        )


class SlashSubCommand(BaseSlashCommand, WithAsOption, WithAsyncCallback, abc.ABC):
    """
    Abstract base class for slash subcommands. All slash subcommands should inherit from this class.

    All abstract methods **must** be implemented by your custom slash subcommand class. A list of the abstract
    methods and properties you are required to implement for this class can be seen below:

    - :obj:`~lightbulb.slash_commands.BaseSlashCommand.description` (class variable)
    - :obj:`~lightbulb.slash_commands.WithAsyncCallback.callback` (instance method)
    """

    __slots__ = ()

    async def __call__(self, *args, **kwargs):
        return await self.callback(*args, **kwargs)

    @functools.lru_cache
    def as_option(self) -> hikari.CommandOption:
        hk_options = _get_options_for_command_instance(self)
        return hikari.CommandOption(
            name=self.name,
            description=self.description,
            type=hikari.OptionType.SUB_COMMAND,
            options=hk_options[::-1],
            is_required=False,
        )



# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__: typing.Final[typing.List[str]] = ["Bot", "when_mentioned_or"]

import asyncio
import collections
import functools
import importlib
import inspect
import logging
import re
import sys
import typing

import hikari
from hikari.internal import ux
from multidict import CIMultiDict

from lightbulb import commands
from lightbulb import context as context_
from lightbulb import errors
from lightbulb import events
from lightbulb import help as help_
from lightbulb import plugins
from lightbulb import slash_commands
from lightbulb.converters import _DefaultingConverter
from lightbulb.converters import _GreedyConverter
from lightbulb.slash_commands.commands import _serialise_command
from lightbulb.utils import maybe_await

_LOGGER = logging.getLogger("lightbulb")

# XXX: can't we use `str.split()` here, which splits on all whitespace in the same way?
ARG_SEP_REGEX = re.compile(r"(?:\s+|\n)")


def when_mentioned_or(prefix_provider):
    """
    Helper function which allows the bot's mentions to be used as the command prefix, as well
    as any other prefix(es) passed in or supplied by the ``prefix_provider``.

    Args:
        prefix_provider: An object that is a prefix, contains prefixes, or returns prefixes.

    Example:

        .. code-block:: python

            # The below are all valid
            bot = lightbulb.Bot(prefix=lightbulb.when_mentioned_or("!"), ...)

            bot = lightbulb.Bot(prefix=lightbulb.when_mentioned_or(["!", "?"]), ...)

            # Using only mentions as the prefix
            bot = lightbulb.Bot(prefix=lightbulb.when_mentioned_or(None), ...)

            # Using with a get_prefix function
            def get_prefix(bot, message):
                # Do something to get the prefixes
                return prefixes

            bot = lightbulb.Bot(prefix=lightbulb.when_mentioned_or(get_prefix), ...)
    """

    async def get_prefixes(bot, message):
        mentions = [f"<@{bot.get_me().id}> ", f"<@!{bot.get_me().id}> "]

        if callable(prefix_provider):
            prefixes = prefix_provider(bot, message)
            if inspect.iscoroutine(prefixes) or isinstance(prefixes, asyncio.Future):
                prefixes = await prefixes
        else:
            prefixes = prefix_provider

        if isinstance(prefixes, str):
            return mentions + [prefixes]
        elif isinstance(prefixes, typing.Iterable):
            return mentions + list(prefixes)
        elif prefixes is None:
            return mentions
        else:
            return mentions + [prefix async for prefix in prefixes]

    return get_prefixes


# Prefixes may be a string or iterable of strings. A string is a sequence of strings too by definition,
# so this type hint _is_ correct.
def _return_prefix(_: typing.Any, __: typing.Any, *, prefixes: typing.Iterable[str]) -> typing.Sequence[str]:
    return prefixes


class Bot(hikari.GatewayBot):
    """
    A subclassed implementation of :class:`hikari.impl.bot.BotAppImpl` which contains a command handler.
    This should be instantiated instead of the superclass if you want to be able to use
    the command handler implementation provided.

    The prefix argument will accept any of the following:

    - A single string, eg ``'!'``

    - An iterable (such as a list) of strings, eg ``['!', '?']``

    - A function or coroutine that takes **only** two arguments, ``bot`` and ``message``, and that returns a single string or iterable of strings.

    Args:
        prefix: The bot's command prefix, iterable of prefixes, or callable that returns
            a prefix or iterable of prefixes.
        insensitive_commands (:obj:`bool`): Whether or not commands should be case-insensitive or not.
            Defaults to False (commands are case-sensitive).
        ignore_bots (:obj:`bool`): Ignore other bot's messages invoking your bot's commands if True (default), else not.
            invoked by other bots. Defaults to ``True``.
        owner_ids (List[ :obj:`int` ]): IDs that the bot should treat as owning the bot.
        help_class (:obj:`~.help.HelpCommand`): The **uninstantiated** class the bot should use for it's help command.
            Defaults to :obj:`~.help.HelpCommand`. Any class passed should always be this class or subclass of
            this class.
        delete_unbound_slash_commands (:obj:`bool`): Whether or not to delete unbound slash commands when the
            bot starts. This will remove any slash commands that do not have a
            :obj:`~lightbulb.slash_commands.SlashCommand` object bound to them when the bot starts but are registered
            according to  discord's API. Defaults to ``True``.
        recreate_changed_slash_commands (:obj:`bool`): Whether or not to send the new version of the slash command
            to discord if the bot detects that the version on discord does not match the local version. Defaults
            to ``True``.
        slash_commands_only (:obj:`bool`): Whether or not the bot will only be using slash commands to interact
            with discord. Defaults to ``False``. If this is ``False`` and no prefix is provided then an error will
            be raised. If ``True``, then you do not need to provide a prefix.
        **kwargs: Other parameters passed to the :class:`hikari.impl.bot.BotAppImpl` constructor.
    """

    def __init__(
        self,
        *,
        prefix=None,
        insensitive_commands: bool = False,
        ignore_bots: bool = True,
        owner_ids: typing.Iterable[int] = (),
        help_class: typing.Type[help_.HelpCommand] = help_.HelpCommand,
        delete_unbound_slash_commands: bool = True,
        recreate_changed_slash_commands: bool = True,
        slash_commands_only: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        if prefix is None and slash_commands_only is False:
            raise TypeError("slash_commands_only is False but no prefix was provided.")

        if not slash_commands_only:
            self.subscribe(hikari.MessageCreateEvent, self.handle)
        self.subscribe(hikari.InteractionCreateEvent, self.handle_slash_commands)
        self.subscribe(hikari.StartingEvent, self._manage_slash_commands)

        if isinstance(prefix, str):
            self.get_prefix = functools.partial(_return_prefix, prefixes=[prefix])
        elif isinstance(prefix, collections.abc.Iterable):
            self.get_prefix = functools.partial(_return_prefix, prefixes=prefix)
        else:
            self.get_prefix = prefix

        self.ignore_bots: bool = ignore_bots
        self.owner_ids: typing.Iterable[int] = owner_ids
        """Iterable of the bot's owner IDs. This can be set by :meth:`Bot.fetch_owner_ids` if not given in the constructor."""
        self.insensitive_commands = insensitive_commands
        self.extensions: typing.List[str] = []
        """A list of extensions currently loaded to the bot."""
        self.plugins: typing.MutableMapping[str, plugins.Plugin] = {}
        """A mapping of plugin name to plugin object currently added to the bot."""
        self._commands: typing.MutableMapping[str, commands.Command] = (
            dict() if not self.insensitive_commands else CIMultiDict()
        )
        self.commands: typing.Set[commands.Command] = set()
        """A set containing all commands and groups registered to the bot."""
        self._checks = []

        self._delete_unbound_slash_commands = delete_unbound_slash_commands
        self._recreate_changed_slash_commands = recreate_changed_slash_commands
        self._slash_commands: typing.MutableMapping[
            str, typing.Union[slash_commands.SlashCommand, slash_commands.SlashCommandGroup]
        ] = {}
        self.slash_commands: typing.Set[
            typing.Union[slash_commands.SlashCommand, slash_commands.SlashCommandGroup]
        ] = set()
        """A set containing all slash commands registered to the bot."""

        self._app: typing.Optional[hikari.PartialApplication] = None

        self._help_impl = help_class(self)

    @staticmethod
    def print_banner(banner: typing.Optional[str], allow_color: bool, force_color: bool) -> None:
        ux.print_banner(banner, allow_color, force_color)
        if banner == "hikari":
            sys.stdout.write(f"Thank you for using lightbulb!\n")

    @property
    def help_command(self):
        """The instance of the help class used by the bot."""
        return self._help_impl

    @help_command.setter
    def help_command(self, new_help_instance: help_.HelpCommand):
        self._help_impl = new_help_instance

    async def fetch_owner_ids(self) -> None:
        """
        Fetches the IDs of the bot's owner(s) from the API and stores them in
        :attr:`~.command_handler.Bot.owner_ids`

        Returns:
            ``None``
        """
        # TODO - add way to flush the owner_id cache
        application = await self.rest.fetch_application()
        self.owner_ids = []
        if application.owner is not None:
            self.owner_ids.append(application.owner.id)
        if application.team is not None:
            self.owner_ids.extend([member_id for member_id in application.team.members])

    def command(self, **kwargs) -> typing.Callable:
        """
        A decorator that registers a coroutine as a command for the handler.

        Keyword Args:
            **kwargs: See :obj:`~.commands.command` for valid kwargs.

        Example:

            .. code-block:: python

                bot = lightbulb.Bot(...)

                @bot.command()
                async def ping(ctx):
                    await ctx.respond("Pong!")

        See Also:
            :meth:`~.command_handler.Bot.add_command`
        """

        def decorate(func: typing.Callable):
            return self.add_command(func, **kwargs)

        return decorate

    def group(self, **kwargs) -> typing.Callable:
        """
        A decorator that registers a coroutine as a command group for the handler.

        Keyword Args:
            **kwargs: See :obj:`~.commands.group` for valid kwargs.

        Example:

            .. code-block:: python

                bot = lightbulb.Bot(...)

                @bot.group()
                async def foo(ctx):
                    await ctx.respond("Bar")

        See Also:
            :meth:`~.commands.Group.command` for how to add subcommands to a group.
        """

        def decorate(func: typing.Callable):
            return self.add_group(func, **kwargs)

        return decorate

    def check(self):
        """
        A decorator that registers a coroutine as a global check.

        This check coroutine will be called before any command invocation.

        Example:

            .. code-block:: python

                bot = lightbulb.Bot(...)

                @bot.check()
                async def global_guild_only(ctx):
                    return ctx.guild_id is not None

        See Also:
            :meth:`~.command_handler.Bot.add_check`
        """

        def decorate(func) -> None:
            self.add_check(func)

        return decorate

    def add_command(self, func: typing.Union[typing.Callable, commands.Command], **kwargs) -> commands.Command:
        """
        Adds a command to the bot. Similar to the ``command`` decorator.

        Args:
            func (:obj:`typing.Callable`): The function or command object to register as a command.

        Keyword Args:
            **kwargs: See :obj:`~.commands.command` for valid kwargs.

        Returns:
            :obj:`~.commands.Command`: Command added to the bot.

        Raises:
            :obj:`NameError`: If the name or an alias of the command being added conflicts with
                an existing command.

        Example:

            .. code-block:: python

                import lightbulb

                bot = lightbulb.Bot(...)

                async def ping(ctx):
                    await ctx.respond("Pong!")

                bot.add_command(ping)

        See Also:
            :meth:`~.command_handler.Bot.command`
        """
        if not isinstance(func, commands.Command):
            name = kwargs.get("name", func.__name__)
            cls = kwargs.get("cls", commands.Command)
            func = cls(
                func,
                name,
                kwargs.get("allow_extra_arguments", True),
                kwargs.get("aliases", []),
                kwargs.get("hidden", False),
            )

        if self.insensitive_commands:
            if set([name.casefold() for name in self._commands.keys()]).intersection(
                {func.name.casefold(), *[a.casefold() for a in func._aliases]}
            ):
                raise NameError(f"Command {func.name!r} has name or alias already registered.")
        else:
            if set(self._commands.keys()).intersection({func.name, *func._aliases}):
                raise NameError(f"Command {func.name!r} has name or alias already registered.")

        self.commands.add(func)
        self._commands[func.name] = func
        for alias in func._aliases:
            self._commands[alias] = func
        _LOGGER.debug("new command registered %r", func.name)
        return self._commands[func.name]

    def add_group(self, func: typing.Union[typing.Callable, commands.Group], **kwargs) -> commands.Group:
        """
        Adds a command group to the bot. Similar to the ``group`` decorator.

        Args:
            func (Union[ :obj:`typing.Callable`, :obj:`~.commands.Group` ]): The function or group object
                to register as a command group.

        Keyword Args:
            **kwargs: See :obj:`~.commands.group` for valid kwargs.

        Returns:
            :obj:`~.commands.Group`: Group added to the bot.

        Raises:
            :obj:`AttributeError`: If the name or an alias of the group being added conflicts with
                an existing command.

        See Also:
            :meth:`~.command_handler.Bot.group`
            :meth:`~.command_handler.Bot.add_command`
        """
        if not isinstance(func, commands.Group):
            name = kwargs.get("name", func.__name__)
            cls = kwargs.get("cls", commands.Group)
            func = cls(
                func,
                name,
                kwargs.get("allow_extra_arguments", True),
                kwargs.get("aliases", []),
                kwargs.get("hidden", False),
                insensitive_commands=kwargs.get("insensitive_commands", False),
            )

        if set(self._commands.keys()).intersection({func.name, *func._aliases}):
            raise AttributeError(f"Command {func.name!r} has name or alias already registered.")

        self.commands.add(func)
        self._commands[func.name] = func
        for alias in func._aliases:
            self._commands[alias] = func
        _LOGGER.debug("new group registered %r", func.name)
        return self._commands[func.name]

    def add_check(self, func: typing.Callable[[context_.Context], typing.Coroutine[None, None, bool]]) -> None:
        """
        Add a coroutine as a global check for the bot.

        This coroutine will be called before any command invocation.

        Args:
            func (Callable[[ :obj:`~.context.Context ], Coroutine[ ``None``, ``None``, :obj:`bool` ]]): The
                coroutine to add as a global check.

        Returns:
            ``None``
        """
        self._checks.append(func)

    def add_plugin(self, plugin: typing.Union[plugins.Plugin, typing.Type[plugins.Plugin]]) -> None:
        """
        Add a :obj:`~.plugins.Plugin` instance or class to the bot, including all of the plugin commands.

        If a class is passed in instead of a plugin instance, then the bot will attempt to instantiate it and
        then add the newly-created plugin instance to the bot.

        Args:
            plugin (Union[:obj:`~.plugins.Plugin`, Type[:obj:`~.plugins.Plugin`]]): Plugin instance or plugin
                subclass to add to the bot.

        Returns:
            ``None``

        Raises:
            :obj:`NameError`: If the name of the plugin being added conflicts with an
                existing plugin.
        """
        if inspect.isclass(plugin) and issubclass(plugin, plugins.Plugin):
            return self.add_plugin(plugin())

        if plugin.name in self.plugins:
            raise NameError(f"A plugin named {plugin.name!r} is already registered.")

        self.plugins[plugin.name] = plugin
        for command in plugin._commands.values():
            if isinstance(command, commands.Group):
                self.add_group(command)
            else:
                self.add_command(command)

        for event_type, listeners in plugin.listeners.items():
            for listener in listeners:
                callback = listener.__get__(plugin, type(plugin))
                self.subscribe(listener.event_type, callback)
                _LOGGER.debug("new listener registered %r (%s)", callback.__name__, listener.event_type.__name__)

        _LOGGER.debug("new plugin registered %r", plugin.name)

    def get_command(self, name: str) -> typing.Optional[commands.Command]:
        """
        Get a command object from its registered name.

        Args:
            name (:obj:`str`): The name of the command to get the object for.

        Returns:
            Optional[ :obj:`~.commands.Command` ]: Command object registered to that name.
        """
        tokens = name.split()
        this = self._commands.get(tokens.pop(0))

        if not tokens:
            return this
        if this is None:
            return this

        for token in tokens:
            this = this.get_subcommand(token)

        return this

    def get_plugin(self, name: str) -> typing.Optional[plugins.Plugin]:
        """
        Get a plugin object from its registered name.

        Args:
             name (:obj:`str`): The name of the plugin to get the object for.

        Returns:
            Optional[ :obj:`~.commands.Command` ]: Plugin object registered to that name.
        """
        return self.plugins.get(name)

    def remove_command(self, name: str) -> typing.Optional[str]:
        """
        Remove a command from the bot and return its name or ``None`` if no command was removed.

        Args:
            name (:obj:`str`): The name of the command to remove.

        Returns:
            Optional[ :obj:`str` ]: Name of the command that was removed.
        """
        command = self._commands.pop(name, None)

        if command is None:
            return None

        self.commands.remove(command)

        keys_to_remove = [command.name, *command._aliases]
        keys_to_remove.remove(name)
        for key in keys_to_remove:
            self._commands.pop(key)
        _LOGGER.debug("command removed %r", command.name)

        return command.name

    def remove_plugin(self, name: str) -> typing.Optional[str]:
        """
        Remove a plugin from the bot and return its name or ``None`` if no plugin was removed.

        Args:
            name (:obj:`str`): The name of the plugin to remove.

        Returns:
            Optional[ :obj:`str` ]: Name of the plugin that was removed.
        """
        plugin = self.plugins.pop(name, None)

        if plugin is None:
            return None

        plugin.plugin_remove()

        for k in plugin._commands.keys():
            self.remove_command(k)

        for event_type, listeners in plugin.listeners.items():
            for listener in listeners:
                callback = listener.__get__(plugin, type(plugin))
                self.unsubscribe(listener.event_type, callback)
                _LOGGER.debug("listener removed %r (%s)", callback.__name__, listener.event_type.__name__)

        _LOGGER.debug("plugin removed %r", plugin.name)

        return plugin.name

    def load_extension(self, extension: str) -> None:
        """
        Load an external extension into the bot. Extension name follows the format ``<directory>.<filename>``
        The extension **must** contain a function ``load`` which takes a single argument which will be the
        bot instance you are loading the extension into.

        Args:
            extension (:obj:`str`): The name of the extension to load.

        Returns:
            ``None``

        Raises:
            :obj:`~.errors.ExtensionAlreadyLoaded`: If the extension has already been loaded.
            :obj:`~.errors.ExtensionMissingLoad`: If the extension to be loaded does not contain a ``load`` function.

        Example:
            This method is useful when wanting to split your bot up into multiple files.
            An example extension is seen below  - ``example.py`` and would be loaded by calling
            the following: ``bot.load_extension("example")``

            .. code-block:: python

                import lightbulb

                class MyPlugin(lightbulb.Plugin):
                    ...

                def load(bot):
                    bot.add_plugin(MyPlugin())
        """
        if extension in self.extensions:
            raise errors.ExtensionAlreadyLoaded(text=f"Extension {extension!r} is already loaded.")

        module = importlib.import_module(extension)

        if not hasattr(module, "load"):
            raise errors.ExtensionMissingLoad(text=f"Extension {extension!r} is missing a load function")
        else:
            module.load(self)
            self.extensions.append(extension)
            _LOGGER.debug("new extension loaded %r", extension)

    def unload_extension(self, extension: str) -> None:
        """
        Unload an external extension from the bot. This method relies on a function, ``unload``
        existing in the extension which the bot will use to remove all commands and/or plugins
        from the bot.

        Args:
            extension (:obj:`str`): The name of the extension to unload.

        Returns:
            ``None``

        Raises:
            :obj:`~.errors.ExtensionNotLoaded`: If the extension has not been loaded.
            :obj:`~.errors.ExtensionMissingUnload`: If the extension does not contain an ``unload`` function.

        Example:

            .. code-block:: python

                import lightbulb

                class MyPlugin(lightbulb.Plugin):
                    ...

                def load(bot):
                    bot.add_plugin(MyPlugin())

                def unload(bot):
                    bot.remove_plugin("MyPlugin")
        """
        if extension not in self.extensions:
            raise errors.ExtensionNotLoaded(text=f"Extension {extension!r} is not loaded.")

        module = importlib.import_module(extension)

        if not hasattr(module, "unload"):
            raise errors.ExtensionMissingUnload(text=f"Extension {extension!r} is missing an unload function")
        else:
            module.unload(self)
            self.extensions.remove(extension)
            del sys.modules[extension]
            _LOGGER.debug("extension unloaded %r", extension)

    def reload_extension(self, extension: str) -> None:
        """
        Reload a bot extension. This method is atomic and so the bot will
        revert to the previous loaded state if the extension encounters a problem
        during unloading or loading.

        Args:
            extension (:obj:`str`): The name of the extension to be reloaded.

        Returns:
            ``None``
        """
        _LOGGER.debug("reloading extension %r", extension)
        old = sys.modules[extension]
        try:
            self.unload_extension(extension)
            self.load_extension(extension)
        except Exception as e:
            sys.modules[extension] = old
            self.load_extension(extension)
            raise e
        else:
            del old

    def walk_commands(self) -> typing.Generator[commands.Command, None, None]:
        """
        A generator that walks through all commands and subcommands registered to the bot.

        Yields:
            :obj:`~.commands.Command`: All commands, groups and subcommands registered to the bot.
        """
        for command in self.commands:
            yield command
            if isinstance(command, commands.Group):
                yield from command.walk_commands()

    async def send_help(
        self, context: context_.Context, obj: typing.Union[commands.Command, plugins.Plugin] = None
    ) -> None:
        """
        Send help to the provided context to the specified object, or send the bot's help overview if
        no object to send help for is supplied.

        Args:
            context (:obj:`~.context.Context`): The context to send help to.
            obj (Union[ :obj:`~.commands.Command`, :obj:`~.plugins.Plugin` ]): The object to send help for.
                Defaults to ``None``.

        Returns:
            ``None``
        """
        if obj is None:
            await self.help_command.send_help_overview(context)
        elif isinstance(obj, commands.Group):
            await self.help_command.send_group_help(context, obj)
        elif isinstance(obj, commands.Command):
            await self.help_command.send_command_help(context, obj)
        elif isinstance(obj, plugins.Plugin):
            await self.help_command.send_plugin_help(context, obj)

    def get_context(
        self, message: hikari.Message, prefix: str, invoked_with: str, invoked_command: commands.Command
    ) -> context_.Context:
        """
        Get the :obj:`~.context.Context` instance for the given arguments. This should be overridden
        if you wish to supply a custom :obj:`~.context.Context` class to your commands.

        Args:
            message (:obj:`hikari.Message`): The message the context is for.
            prefix (:obj:`str`): The prefix used in this context.
            invoked_with (:obj:`str`): The command name/alias used to trigger invocation.
            invoked_command (:obj:`~.commands.Command`): The command that will be invoked.

        Returns:
            :obj:`~.context.Context`: The context to be used for the command invocation.
        """
        return context_.Context(self, message, prefix, invoked_with, invoked_command)

    async def _evaluate_checks(self, command: commands.Command, context: context_.Context) -> bool:
        checks = [*self._checks, *command.checks]
        if command.plugin is not None:
            checks.append(command.plugin.plugin_check)

        failed_checks = []
        for check in checks:
            try:
                if not await check(context):
                    failed_checks.append(
                        errors.CheckFailure(f"Check {check.__name__} failed for command {command.name}")
                    )
            except Exception as ex:
                error = errors.CheckFailure(str(ex))
                error.__cause__ = ex
                failed_checks.append(ex)

        if len(failed_checks) > 1:
            raise errors.CheckFailure("Multiple checks failed: " + ", ".join(str(ex) for ex in failed_checks))
        elif failed_checks:
            raise failed_checks[0]
        return True

    async def _dispatch_command_error_event_from_exception(
        self,
        exception: errors.CommandError,
        message: hikari.Message,
        context: typing.Optional[context_.Context] = None,
        command: typing.Optional[commands.Command] = None,
    ) -> None:
        error_event = events.CommandErrorEvent(
            app=self, exception=exception, message=message, context=context, command=command
        )

        handled = False
        if command is not None:
            if command._error_listener is not None:
                handled = bool(await command._error_listener(error_event))
            if not handled and self.get_listeners(events.CommandErrorEvent, polymorphic=True):
                await self.dispatch(error_event)
                handled = True
        else:
            if self.get_listeners(events.CommandErrorEvent, polymorphic=True):
                await self.dispatch(error_event)
                handled = True

        if not handled:
            raise exception

    async def _resolve_prefix(self, message: hikari.Message) -> typing.Optional[str]:
        prefixes = await maybe_await(self.get_prefix, self, message)

        if isinstance(prefixes, str):
            prefixes = [prefixes]

        prefixes.sort(key=len, reverse=True)

        for prefix in prefixes:
            if message.content.startswith(prefix):
                return prefix
        return None

    def _validate_command_exists(self, invoked_with) -> commands.Command:
        if (command := self.get_command(invoked_with)) is not None:
            return command
        raise errors.CommandNotFound(invoked_with)

    async def resolve_args_for_command(
        self, context: context_.Context, command: commands.Command, raw_arg_string: str
    ) -> typing.Tuple[typing.List[str], typing.Dict[str, str]]:
        """
        Resolve the appropriate command arguments from an unparsed string
        containing the raw command arguments and attempt to convert them into
        the appropriate types given the type hint.

        This method can be overridden if you wish to customise how arguments
        are parsed and converted for all the commands. If you override this then it is important
        that it at least returns a tuple containing an empty list if no positional arguments
        were resolved, and an empty dict if no keyword arguments were resolved.

        If you override this method then you may find the :obj:`~.stringview.StringView`
        class useful for extracting the arguments from the raw string and the property
        :obj:`~.command.arg_details` which contains information about the command's arguments
        and which are optional or required.

        Args:
            context (:obj:`~.context.Context`): The invocation context for the command.
            command (:obj:`~.commands.Command`): The command to resolve the arguments for.
            raw_arg_string (:obj:`str`): String containing the raw, unparsed arguments.

        Returns:
            Tuple[ List[ :obj:`str` ], Dict[ :obj:`str1, :obj:`str` ] ]: Positional and keyword
                arguments the command should be invoked with.

        Raises:
            :obj:`~.errors.TooManyArguments`: The command does not ignore extra arguments and too many
                arguments were supplied by the user.
            :obj:`~.errors.NotEnoughArguments`: Not enough arguments were provided by the user to fill
                all required argument fields.
            :obj:`~.errors.ConverterFailure`: Argument value conversion failed.
        """
        converters = command.arg_details.converters[:]
        arg_names = command.arg_details.arguments[:]
        args, kwargs = [], {}

        arg_string = raw_arg_string
        while converters:
            conv = converters.pop(0)
            arg_name = arg_names.pop(0)

            if not isinstance(conv, (_DefaultingConverter, _GreedyConverter)) and not arg_string:
                raise errors.NotEnoughArguments(command, [arg_name, *arg_names])

            try:
                conv_out, arg_string = await conv.convert(context, arg_string)
            except (ValueError, TypeError, errors.ConverterFailure):
                raise errors.ConverterFailure(f"Converting failed for argument {arg_name!r}")

            if isinstance(conv_out, dict):
                kwargs.update(conv_out)
            elif isinstance(conv, _GreedyConverter) and conv.unpack:
                args.extend(conv_out)
            else:
                args.append(conv_out)

        if arg_string and not command._allow_extra_arguments:
            raise errors.TooManyArguments(command)

        return args, kwargs

    async def _invoke_command(
        self,
        command: commands.Command,
        context: context_.Context,
        args: typing.Sequence[str],
        kwargs: typing.Mapping[str, str],
    ) -> None:
        await command.invoke(context, *args, **kwargs)

    async def process_commands_for_event(self, event: hikari.MessageCreateEvent) -> None:
        """
        Carries out all command and argument parsing, evaluates checks and ultimately invokes
        a command if the event passed is deemed to contain a command invocation.

        It is not recommended that you override this method - if you do you should make sure that
        you know what you are doing.

        Args:
            event (:obj:`hikari.MessageCreateEvent`): The event to process commands for.

        Returns:
            ``None``
        """
        prefix = await self._resolve_prefix(event.message)
        if prefix is None:
            return

        new_content = event.message.content[len(prefix) :]

        if not new_content or new_content.isspace():
            return

        split_args = ARG_SEP_REGEX.split(new_content, maxsplit=1)
        invoked_with, command_args = split_args[0], "".join(split_args[1:])

        if not invoked_with:
            # Return if the character immediately following the command prefix
            # is whitespace to prevent IndexError later on
            return

        try:
            command = self._validate_command_exists(invoked_with)
        except errors.CommandNotFound as ex:
            await self._dispatch_command_error_event_from_exception(ex, event.message)
            return

        temp_args = command_args
        final_args = command_args
        while isinstance(command, commands.Group) and command_args:
            next_split = ARG_SEP_REGEX.split(temp_args, maxsplit=1)
            next_arg, temp_args = next_split[0], "".join(next_split[1:])
            prev_command = command
            maybe_subcommand = command.get_subcommand(next_arg)

            if maybe_subcommand is None:
                command = prev_command
                break
            else:
                command = maybe_subcommand
                final_args = temp_args

        context = self.get_context(event.message, prefix, invoked_with, command)

        await self.dispatch(events.CommandInvocationEvent(app=self, command=command, context=context))
        if (before_invoke := command._before_invoke) is not None:
            await before_invoke(context)

        try:
            positional_args, keyword_arg = await self.resolve_args_for_command(context, command, final_args)
            if not await maybe_await(command._check_exempt_predicate, context):
                await self._evaluate_checks(command, context)
            else:
                _LOGGER.debug("checks bypassed for command %r", context.message.content)
        except (
            errors.NotEnoughArguments,
            errors.TooManyArguments,
            errors.CheckFailure,
            errors.CommandSyntaxError,
            errors.ConverterFailure,
        ) as ex:
            await self._dispatch_command_error_event_from_exception(ex, event.message, context, command)
            return

        try:
            await self._invoke_command(command, context, positional_args, keyword_arg)
        except errors.CommandError as ex:
            await self._dispatch_command_error_event_from_exception(ex, event.message, context, command)
            return
        except Exception as ex:
            new_ex = errors.CommandInvocationError(f"{type(ex).__name__}: {ex}", ex)
            await self._dispatch_command_error_event_from_exception(new_ex, event.message, context, command)
            return

        if (after_invoke := command._after_invoke) is not None:
            await after_invoke(context)

        await self.dispatch(events.CommandCompletionEvent(app=self, command=command, context=context))

    async def handle(self, event: hikari.MessageCreateEvent) -> None:
        """
        The message listener that deals with validating the invocation messages. If invocation message
        is valid then it will delegate parsing and invocation to :obj:`Bot.process_commands_for_event`.

        You can override this method to customise how the bot should validate that a message could
        contain a command invocation, for example making it ignore specific guilds or users.

        If you choose to override this method, it should await :obj:`Bot.process_commands_for_event`
        otherwise no commands will ever be invoked.

        Args:
            event (:obj:`hikari.events.message.MessageCreateEvent`): The message create event containing
                a possible command invocation.

        Returns:
            ``None``
        """
        if self.ignore_bots and not event.is_human:
            return

        if not event.message.content:
            return

        await self.process_commands_for_event(event)

    def add_slash_command(self, command: typing.Type[slash_commands.BaseSlashCommand], create: bool = False) -> None:
        """
        Registers a slash command with the bot.

        Args:
            command (Type[:obj:`~lightbulb.slash_commands.SlashCommandBase`]): The slash command class to register
                to the bot. This should **not** be instantiated.
            create (:obj:`bool`): Whether or not to send a create request to discord when the command is added.
                Defaults to ``False``.

        Returns:
            ``None``

        Warning:
            The ``create`` argument **must** be ``False`` if the command is being added to the bot before
            the bot has started (i.e. ``bot.run()`` has not yet been called).
        """
        cmd = command(self)

        if cmd.name in self._slash_commands:
            raise NameError(f"Slash command {cmd.name!r} has a name already registered.")

        self._slash_commands[cmd.name] = cmd
        self.slash_commands.add(cmd)

        if create and self._app is not None:
            _LOGGER.debug("creating slash command %r", cmd.name)
            asyncio.create_task(cmd.auto_create(self._app))
        elif create and self._app is None:
            _LOGGER.debug("not adding slash command %r as the bot has not started", cmd.name)

        _LOGGER.debug("slash command added %r", cmd.name)

    def remove_slash_command(self, name: str, delete: bool = False) -> typing.Optional[str]:
        """
        Remove a slash command from the bot and return its name or ``None`` if no command was removed.

        Args:
            name (:obj:`str`): The name of the slash command to remove.
            delete (:obj:`bool`): Whether or not to delete the command from discord when it is removed. Defaults
                to ``False``.

        Returns:
            Optional[ :obj:`str` ]: Name of the slash command that was removed.
        """
        name = name.lower()

        cmd = self._slash_commands.pop(name)
        if cmd is not None:
            self.slash_commands.remove(cmd)

            if delete and self._app is not None:
                _LOGGER.debug("purging slash command %r", cmd.name)
                asyncio.create_task(cmd.auto_delete(self._app))
            elif delete and self._app is None:
                _LOGGER.debug("not purging slash command %r as the bot has not started", cmd.name)

            _LOGGER.debug("slash command removed %r", cmd.name)

        return cmd.name if cmd is not None else None

    def get_slash_command(
        self, name: str
    ) -> typing.Optional[typing.Union[slash_commands.SlashCommand, slash_commands.SlashCommandGroup]]:
        """
        Gets the slash command with the given name, or ``None`` if one with that name does
        not exist.

        Args:
            name (:obj:`str`): The name of the slash command to get the object for.

        Returns:
            Optional[:obj:`~lightbulb.slash_commands.BaseSlashCommand`]: Retrieved slash command or ``None``
                if not found.
        """
        return self._slash_commands.get(name)

    async def purge_slash_commands(self, *guild_ids: hikari.Snowflakeish, global_commands: bool = False) -> None:
        """
        Purges all slash commands from the guilds with the specified IDs, and all the global slash
        commands if ``global_commands`` is ``True``. Useful if you want to teardown old slash commands from
        the bot and cannot be bothered to fetch them individually yourself. If neither `guild_ids` nor `global_commands`
        is specified then this method will do nothing.

        Args:
            *guild_ids (:obj:`hikari.Snowflakeish`): IDs for the guilds to purge slash commands from.

        Keyword Args:
            global_commands (:obj:`bool`): Whether or not to purge global slash commands from the bot.

        Returns:
            ``None``
        """
        commands_to_remove = []
        if global_commands:
            commands_to_remove.extend(await self.rest.fetch_application_commands(self._app))
        if guild_ids:
            for guild_id in guild_ids:
                commands_to_remove.extend(await self.rest.fetch_application_commands(self._app, guild_id))

        for command in commands_to_remove:
            if command.guild_id is None:
                _LOGGER.debug("deleting global slash command %r", command.name)
            else:
                _LOGGER.debug("deleting slash command %r from guild %r", command.name, str(command.guild_id))
            await command.delete()

    async def handle_slash_commands(self, event: hikari.InteractionCreateEvent) -> None:
        """
        The InteractionCreateEvent listener that handles slash command invocations and
        resolves and invokes the correct callback from the event payload.

        Args:
            event (:obj:`hikari.InteractionCreateEvent`): The InteractionCreateEvent for the slash command
                invocation.
        Returns:
            ``None``
        """
        if not isinstance(event.interaction, hikari.CommandInteraction):
            return

        command = self.get_slash_command(event.interaction.command_name)
        if command is None:
            return

        context = slash_commands.SlashCommandContext(self, event.interaction, command)
        _LOGGER.debug("invoking slash command %r", command.name)
        await command(context)

    async def _manage_slash_commands(self, _):
        self._app = await self.rest.fetch_application()

        global_slash_commands = {c.name: c for c in await self.rest.fetch_application_commands(self._app)}

        guilds = set()
        for cmd in self._slash_commands.values():
            if cmd.enabled_guilds is not None:
                guilds.update(cmd.enabled_guilds)

        _guild_slash_commands = {
            g_id: {c.name: c for c in await self.rest.fetch_application_commands(self._app, g_id)} for g_id in guilds
        }
        guild_slash_commands = collections.defaultdict(list)
        for g_id, cmd_info in _guild_slash_commands.items():
            for cmd in cmd_info.values():
                guild_slash_commands[cmd.name].append([cmd, g_id])

        # Note that the purge_slash_commands method is not used here as the bot only purges
        # commands that appear not to have an implementation registered, as opposed to just
        # purging every command.
        remaining_globals = {} if self._delete_unbound_slash_commands is True else global_slash_commands
        if self._delete_unbound_slash_commands:
            _LOGGER.debug("purging unbound slash commands")
            for cmd in global_slash_commands.values():
                # if an implementation of the slash command is not found
                if self._slash_commands.get(cmd.name) is None:
                    _LOGGER.debug("deleting global slash command %r", cmd.name)
                    await cmd.delete()
                # if our implementation of the slash command is specific to guilds
                elif self._slash_commands[cmd.name].enabled_guilds is not None:
                    _LOGGER.debug("deleting global slash command %r", cmd.name)
                    await cmd.delete()
                else:
                    remaining_globals[cmd.name] = cmd

        remaining_guild_cmds = {}
        if self._delete_unbound_slash_commands:
            for cmd_name, cmds in guild_slash_commands.items():
                for cmd, guild_id in cmds:
                    # if an implementation of the slash command is not found
                    if self._slash_commands.get(cmd.name) is None:
                        _LOGGER.debug("deleting slash command %r from guild %r", cmd.name, str(cmd.guild_id))
                        await cmd.delete()
                    # if our implementation of the slash command doesn't contain an entry for this guild
                    elif cmd.guild_id not in self._slash_commands[cmd.name].enabled_guilds:
                        _LOGGER.debug("deleting slash command %r from guild %r", cmd.name, str(cmd.guild_id))
                        await cmd.delete()
                    else:
                        # We are assuming here that all the guild commands with the same name have the same
                        # implementation so that we don't have to do a million more checks
                        if cmd.name in remaining_guild_cmds:
                            remaining_guild_cmds[cmd.name][1].append(cmd.guild_id)
                        else:
                            remaining_guild_cmds[cmd.name] = [cmd, [cmd.guild_id]]
        else:
            for cmd_name, cmds in guild_slash_commands.items():
                for cmd, guild_id in cmds:
                    if cmd.name in remaining_guild_cmds:
                        remaining_guild_cmds[cmd.name][1].append(guild_id)
                    else:
                        remaining_guild_cmds[cmd.name] = [cmd, [guild_id]]

        def compare_commands(
            lb_cmd: typing.Union[slash_commands.SlashCommand, slash_commands.SlashCommandGroup],
            hk_cmd: hikari.Command,
            _guild_ids: typing.Optional[typing.List] = None,
        ) -> bool:
            # If one command is global and the other isn't
            if lb_cmd.enabled_guilds != _guild_ids and (lb_cmd.enabled_guilds is None or _guild_ids is None):
                return False

            # If both commands are global
            if _guild_ids is None and lb_cmd.enabled_guilds is None:
                return _serialise_command(lb_cmd) == _serialise_command(hk_cmd)

            # If both commands are guild commands and for the same guilds
            if set(lb_cmd.enabled_guilds) == set(_guild_ids):
                return _serialise_command(lb_cmd) == _serialise_command(hk_cmd)

            # None of the above
            return False

        # Check if existing slash commands have changed before creating any. If a command does not appear
        # to have changed then we don't send the api request to discord to prevent unnecessary rate-limiting and
        # to prevent the 1h wait period when creating global commands
        if self._recreate_changed_slash_commands:
            for cmd in remaining_globals.values():
                if cmd.name not in self._slash_commands:
                    continue

                if not compare_commands(self._slash_commands[cmd.name], cmd):
                    _LOGGER.debug("recreating global slash command %r as it appears to have changed", cmd.name)
                    await self._slash_commands[cmd.name].auto_create(self._app)
                else:
                    _LOGGER.debug(
                        "not recreating global slash command %r as it doesn't appear to have changed", cmd.name
                    )

            for cmd, guild_ids in remaining_guild_cmds.values():
                if cmd.name not in self._slash_commands:
                    continue

                if not compare_commands(self._slash_commands[cmd.name], cmd, guild_ids):
                    _LOGGER.debug("recreating guild slash command %r as it appears to have changed", cmd.name)
                    await self._slash_commands[cmd.name].auto_create(self._app)
                else:
                    _LOGGER.debug(
                        "not recreating guild slash command %r as it doesn't appear to have changed", cmd.name
                    )

        all_cmd_names = [
            *[c.name for c in remaining_globals.values()],
            *[c.name for c, _ in remaining_guild_cmds.values()],
        ]
        for cmd_name, cmd in self._slash_commands.items():
            if cmd_name not in all_cmd_names:
                _LOGGER.debug("creating slash command %r as it does not seem to exist yet", cmd_name)
                await cmd.auto_create(self._app)



# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
"""
.. _converters:

Lightbulb will attempt to convert command arguments using the type or function specified by the type
hints of each argument. By default, any types that can be converted to from a string are supported - eg int or
float.

Example:
::

    @bot.command()
    async def to_int(ctx, number: int):
        await ctx.respond(f"{number}, {type(number)}")
        # If ran with the argument "10"
        # The output would be: 10, <class 'int'>

Supplied in this module are further utility functions that can be specified as type hints in order
to convert arguments into more complex types - eg member or channel objects.

You can also write your own converters. A converter can be any callable and should take a single argument, which
will be an instance of the :obj:`~lightbulb.converters.WrappedArg` class. The arg value and command invocation
context can be accessed through this instance from the attributes ``data`` and ``context`` respectively.

Hikari classes are also available as type hints in place of the lightbulb converters and will be internally
converted into the necessary converter for the command to behave as expected. A list of all available classes
along with the converters they 'replace' can be seen below:

- :obj:`hikari.User` (:obj:`~.user_converter`)
- :obj:`hikari.Member` (:obj:`~.member_converter`)
- :obj:`hikari.TextableChannel` (:obj:`~.text_channel_converter`)
- :obj:`hikari.GuildVoiceChannel` (:obj:`~.guild_voice_channel_converter`)
- :obj:`hikari.GuildCategory` (:obj:`~.category_converter`)
- :obj:`hikari.Role` (:obj:`~.role_converter`)
- :obj:`hikari.Emoji` (:obj:`~.emoji_converter`)
- :obj:`hikari.GuildPreview` (:obj:`~.guild_converter`)
- :obj:`hikari.Message` (:obj:`~.message_converter`)
- :obj:`hikari.Invite` (:obj:`~.invite_converter`)
- :obj:`hikari.Colour` (:obj:`~.colour_converter`)
- :obj:`hikari.Color` (:obj:`~.color_converter`)


.. warning:: For the supplied converters, some functionality will not be possible depending on the intents and/or
    cache settings of your bot application and object. If the bot does not have a cache then the converters can
    only work for arguments of ID or mention and **not** any form of name.

.. warning:: If you use ``from __future__ import annotations`` then you **will not** be able to use converters
    in your commands. Instead of converting the arguments, the raw, unconverted arguments will be passed back
    to the command.
"""
from __future__ import annotations

__all__: typing.Final[typing.List[str]] = [
    "WrappedArg",
    "user_converter",
    "member_converter",
    "text_channel_converter",
    "guild_voice_channel_converter",
    "category_converter",
    "role_converter",
    "emoji_converter",
    "guild_converter",
    "message_converter",
    "invite_converter",
    "colour_converter",
    "color_converter",
    "Greedy",
]

import collections
import re
import typing
import warnings

import hikari

from lightbulb import context as context_
from lightbulb import errors
from lightbulb import stringview
from lightbulb import utils

T = typing.TypeVar("T")
T_co = typing.TypeVar("T_co", covariant=True)

USER_MENTION_REGEX: typing.Final[typing.Pattern] = re.compile(r"<@!?(\d+)>")
CHANNEL_MENTION_REGEX: typing.Final[typing.Pattern] = re.compile(r"<#(\d+)>")
ROLE_MENTION_REGEX: typing.Final[typing.Pattern] = re.compile(r"<@&(\d+)>")
EMOJI_MENTION_REGEX: typing.Final[typing.Pattern] = re.compile(r"<a?:\w+:(\d+)>")


class WrappedArg(collections.UserString):
    """
    A wrapped command argument containing the invocation context
    of the command for which the argument is to be converted under, accessible
    through the :attr:`context` attribute.

    This class acts like a string so any operations that will work on a string will
    also work on this class.

    The raw string argument can be accessed through the :attr:`data` attribute.

    Args:
        seq (:obj:`str`): The argument text.
        context (:obj:`~.context.Context`): The command invocation context for the argument.
    """

    __slots__ = ["context"]

    def __init__(self, seq: str, context: context_.Context) -> None:
        super().__init__(seq)
        self.context: context_.Context = context

    def __iter__(self):
        return iter(self.data)


def _resolve_id_from_arg(arg_string: str, regex: typing.Pattern) -> hikari.Snowflake:
    if match := regex.match(arg_string):
        arg_string = match.group(1)
    return hikari.Snowflake(arg_string)


async def _get_or_fetch_guild_channel_from_id(arg: WrappedArg, channel_id: hikari.Snowflake) -> hikari.GuildChannel:
    channel = arg.context.bot.cache.get_guild_channel(channel_id)
    if channel is None:
        channel = await arg.context.bot.rest.fetch_channel(channel_id)
    return channel


def _raise_if_not_none(obj: typing.Optional[T]) -> T:
    if obj is None:
        raise errors.ConverterFailure
    return obj


async def user_converter(arg: WrappedArg) -> hikari.User:
    """
    Converter to transform a command argument into a :obj:`hikari.UserImpl` object.

    Args:
        arg (:obj:`WrappedArg`): Argument to transform.

    Returns:
        :obj:`hikari.UserImpl`: The user object resolved from the argument.

    Raises:
        :obj:`~.errors.ConverterFailure`: If the argument could not be resolved into a user object.

    Example:

        .. code-block:: python

            @bot.command()
            async def username(ctx, user: lightbulb.user_converter):
                await ctx.respond(user.username)
    """
    try:
        user_id = _resolve_id_from_arg(arg.data, USER_MENTION_REGEX)
    except ValueError:
        users = arg.context.bot.cache.get_users_view()
        user = utils.find(
            users.values(), lambda u: u.username == arg.data or f"{u.username}#{u.discriminator}" == arg.data
        )
    else:
        user = arg.context.bot.cache.get_user(user_id)
        if user is None:
            user = await arg.context.bot.rest.fetch_user(user_id)
    return _raise_if_not_none(user)


async def member_converter(arg: WrappedArg) -> hikari.Member:
    """
    Converter to transform a command argument into a :obj:`~hikari.Member` object.

    Args:
        arg (:obj:`WrappedArg`): Argument to transform.

    Returns:
        :obj:`~hikari.Member`: The member object resolved from the argument.

    Raises:
        :obj:`~.errors.ConverterFailure`: If the argument could not be resolved into a member object.
    """
    try:
        user_id = _resolve_id_from_arg(arg.data, USER_MENTION_REGEX)
    except ValueError:
        members = arg.context.bot.cache.get_members_view_for_guild(arg.context.guild_id)
        member = utils.find(
            members.values(),
            lambda m: m.username == arg.data or m.nickname == arg.data or f"{m.username}#{m.discriminator}" == arg.data,
        )
    else:
        member = arg.context.bot.cache.get_member(arg.context.guild_id, user_id)
        if member is None:
            member = await arg.context.bot.rest.fetch_member(arg.context.guild_id, user_id)
    return _raise_if_not_none(member)


async def text_channel_converter(arg: WrappedArg) -> hikari.TextableChannel:
    """
    Converter to transform a command argument into a :obj:`~hikari.TextableChannel` object.

    Args:
        arg (:obj:`WrappedArg`): Argument to transform.

    Returns:
        :obj:`hikari.TextableChannel`: The channel object resolved from the argument.

    Raises:
        :obj:`~.errors.ConverterFailure`: If the argument could not be resolved into a channel object.
    """
    try:
        channel_id = _resolve_id_from_arg(arg.data, CHANNEL_MENTION_REGEX)
    except ValueError:
        channels = arg.context.bot.cache.get_guild_channels_view_for_guild(arg.context.guild_id)
        channel = utils.get(channels.values(), name=arg.data)
    else:
        channel = await _get_or_fetch_guild_channel_from_id(arg, channel_id)

    if not isinstance(channel, hikari.TextableChannel):
        raise errors.ConverterFailure
    return channel


async def guild_voice_channel_converter(arg: WrappedArg) -> hikari.GuildVoiceChannel:
    """
    Converter to transform a command argument into a :obj:`~hikari.GuildVoiceChannel` object.

    Args:
        arg (:obj:`WrappedArg`): Argument to transform.

    Returns:
        :obj:`~hikari.GuildVoiceChannel`: The channel object resolved from the argument.

    Raises:
        :obj:`~.errors.ConverterFailure`: If the argument could not be resolved into a channel object.
    """
    try:
        channel_id = _resolve_id_from_arg(arg.data, CHANNEL_MENTION_REGEX)
    except ValueError:
        channels = arg.context.bot.cache.get_guild_channels_view_for_guild(arg.context.guild_id)
        channel = utils.get(channels.values(), name=arg.data)
    else:
        channel = await _get_or_fetch_guild_channel_from_id(arg, channel_id)

    if not isinstance(channel, hikari.GuildVoiceChannel):
        raise errors.ConverterFailure
    return channel


# Deprecated, as this may change in the future to not just be a GuildVoiceChannel should Discord add new features,
# so this shouldn't be used as it may cause breaking changes later.
def voice_channel_converter(arg: WrappedArg) -> typing.Coroutine[typing.Any, typing.Any, hikari.GuildVoiceChannel]:
    warnings.warn(
        "voice_channel_converter is deprecated, use guild_voice_channel_converter instead",
        category=DeprecationWarning,
    )
    return guild_voice_channel_converter(arg)


async def category_converter(arg: WrappedArg) -> hikari.GuildCategory:
    """
    Converter to transform a command argument into a :obj:`~hikari.GuildCategory` object.

    Args:
        arg (:obj:`WrappedArg`): Argument to transform.

    Returns:
        :obj:`~hikari.GuildCategory`: The category object resolved from the argument.

    Raises:
        :obj:`~.errors.ConverterFailure`: If the argument could not be resolved into a category object.
    """
    try:
        channel_id = _resolve_id_from_arg(arg.data, CHANNEL_MENTION_REGEX)
    except ValueError:
        channels = arg.context.bot.cache.get_guild_channels_view_for_guild(arg.context.guild_id)
        channel = utils.get(channels.values(), name=arg.data)
    else:
        channel = await _get_or_fetch_guild_channel_from_id(arg, channel_id)

    if not isinstance(channel, hikari.GuildCategory):
        raise errors.ConverterFailure
    return channel


async def role_converter(arg: WrappedArg) -> hikari.Role:
    """
    Converter to transform a command argument into a :obj:`~hikari.Role` object.

    Args:
        arg (:obj:`WrappedArg`): Argument to transform.

    Returns:
        :obj:`~hikari.Role`: The role object resolved from the argument.

    Raises:
        :obj:`~.errors.ConverterFailure`: If the argument could not be resolved into a role object.
    """
    try:
        role_id = _resolve_id_from_arg(arg.data, ROLE_MENTION_REGEX)
    except ValueError:
        roles = arg.context.bot.cache.get_roles_view_for_guild(arg.context.guild_id)
        role = utils.get(roles.values(), name=arg.data)
    else:
        role = arg.context.bot.cache.get_role(role_id)
        if role is None:
            roles = await arg.context.bot.rest.fetch_roles(arg.context.guild_id)
            roles = dict([(r.id, r) for r in roles])
            role = roles[role_id]
    return _raise_if_not_none(role)


async def emoji_converter(arg: WrappedArg) -> hikari.Emoji:
    """
    Converter to transform a command argument into a :obj:`~hikari.Emoji` object.

    Note that this **does not** validate unicode emojis to ensure they are defined
    as standard emojis. See https://github.com/nekokatt/hikari/issues/270 for discussion
    on supporting this.

    Args:
        arg (:obj:`WrappedArg`): Argument to transform.

    Returns:
        :obj:`~hikari.Emoji`: The emoji object resolved from the argument.

    Raises:
        :obj:`~.errors.ConverterFailure`: If the argument could not be resolved into an emoji object.
    """
    return hikari.Emoji.parse(arg.data)


async def guild_converter(arg: WrappedArg) -> hikari.GuildPreview:
    """
    Converter to transform a command argument into a :obj:`~hikari.Guild` object.

    Args:
        arg (:obj:`WrappedArg`): Argument to transform.

    Returns:
        :obj:`~hikari.Guild`: The guild object resolved from the argument.

    Raises:
        :obj:`~.errors.ConverterFailure`: If the argument could not be resolved into a guild object.
    """
    if arg.data.isdigit():
        guild_id = int(arg.data)
        guild = await arg.context.bot.rest.fetch_guild_preview(guild_id)
    else:
        guilds = arg.context.bot.cache.get_available_guilds_view()
        guild = utils.get(guilds.values(), name=arg.data)

        guild = await arg.context.bot.rest.fetch_guild_preview(guild.id)

    return _raise_if_not_none(guild)


async def message_converter(arg: WrappedArg) -> hikari.Message:
    """
    Converter to transform a command argument into a :obj:`~hikari.Message` object. Note that
    this converter will only return messages from the same context that the command was invoked from, that
    is to say the channel ID of the command invocation and the fetched message must be the same.

    Args:
        arg (:obj:`WrappedArg`): Argument to transform.

    Returns:
        :obj:`~hikari.Message`: The message object resolved from the argument.

    Raises:
        :obj:`~.errors.ConverterFailure`: If the argument could not be resolved into a message object.
    """
    try:
        message_id, channel_id = int(arg.data), arg.context.channel_id
    except ValueError:
        parts = arg.data.rstrip("/").split("/")
        message_id, channel_id = parts[-1], parts[-2]

    if channel_id != arg.context.channel_id:
        raise errors.ConverterFailure

    return await arg.context.bot.rest.fetch_message(channel_id, message_id)


async def invite_converter(arg: WrappedArg) -> hikari.Invite:
    """
    Converter to transform a command argument into an :obj:`~hikari.Invite` object.

    Args:
        arg (:obj:`WrappedArg`): Argument to transform.

    Returns:
        :obj:`~hikari.Invite`: The invite object resolved from the argument.

    Raises:
        :obj:`~.errors.ConverterFailure`: If the argument could not be resolved into an invite object.
    """
    invite_code = arg.data.rstrip("/").split("/")[-1]
    invite = arg.context.bot.cache.get_invite(invite_code)
    if invite is None:
        invite = await arg.context.bot.rest.fetch_invite(invite_code)
    return invite


async def colour_converter(arg: WrappedArg) -> hikari.Colour:
    """
    Converter to transform a command argument into a :obj:`~hikari.Colour` object.

    Args:
        arg (:obj:`WrappedArg`): Argument to transform.

    Returns:
        :obj:`~hikari.Colour`: The colour object resolved from the argument.

    Raises:
        :obj:`~.errors.ConverterFailure`: If the argument could not be resolved into a colour object.
    """
    return hikari.Colour.of(arg.data)


async def color_converter(arg: WrappedArg) -> hikari.Color:
    """Alias for :obj:`~colour_converter`"""
    return await colour_converter(arg)


class Greedy(typing.Generic[T]):
    """
    A special converter that greedily consumes arguments until it either runs out of arguments, or a parser error
    is encountered. Due to this behaviour, most input errors will be silently ignored.

    Example:

        .. code-block:: python

            @bot.command()
            async def foo(ctx, foo: Greedy[int]):
                # if called with <p>foo 1 2 3 4 5
                # then the arg foo would contain [1, 2, 3, 4, 5]
                ...
    """


class _ConverterT(typing.Protocol[T_co]):
    def __call__(self, arg: WrappedArg) -> typing.Union[T_co, typing.Coroutine[None, None, T_co]]:
        ...


class _BaseConverter(typing.Protocol[T_co]):
    async def convert(self, context: context_.Context, arg_string: str, *, parse: bool) -> typing.Tuple[T_co, str]:
        ...


class _Converter:
    __slots__ = ("conversion_func",)

    def __init__(self, conversion_func: _ConverterT) -> None:
        self.conversion_func = conversion_func

    async def convert(self, context: context_.Context, arg_string: str, *, parse: bool = True) -> typing.Tuple[T, str]:
        args, remainder = arg_string, ""
        if parse:
            sv = stringview.StringView(arg_string)
            parsed, remainder = sv.deconstruct_str(max_parse=1)
            args = " ".join(parsed)

        converted_arg = await utils.maybe_await(self.conversion_func, WrappedArg(args, context))
        return converted_arg, remainder


class _UnionConverter:
    __slots__ = ("converters",)

    def __init__(self, *converters: _BaseConverter[T]) -> None:
        self.converters = converters

    async def convert(self, context: context_.Context, arg_string: str, *, parse: bool = False) -> typing.Tuple[T, str]:
        sv = stringview.StringView(arg_string)
        args, remainder = sv.deconstruct_str(max_parse=1)

        for converter in self.converters:
            try:
                converted_arg = await converter.convert(context, " ".join(args), parse=False)
                break
            except Exception:
                continue
        else:
            raise errors.ConverterFailure

        return converted_arg[0], remainder


class _GreedyConverter:
    __slots__ = ("converter", "unpack")

    def __init__(self, converter: _BaseConverter[T], unpack: bool = False) -> None:
        self.converter = converter
        self.unpack = unpack

    async def convert(
        self, context: context_.Context, arg_string: str, *, parse: bool = False
    ) -> typing.Tuple[typing.List[T], str]:
        prev = arg_string
        sv = stringview.StringView(arg_string)
        converted = []

        while True:
            args, remainder = sv.deconstruct_str(max_parse=1)
            if not args:
                break

            try:
                converted_arg = await self.converter.convert(context, " ".join(args), parse=False)
                converted.append(converted_arg[0])
                prev = remainder
            except Exception:
                break

        return converted, prev


class _DefaultingConverter:
    __slots__ = ("converter", "_default", "raise_on_fail")

    def __init__(self, converter: _BaseConverter[T], default: typing.Any, raise_on_fail: bool = True):
        self.converter = converter
        self.raise_on_fail = raise_on_fail
        self._default = default
        self.apply_default_recursively(self.converter, default)

    @property
    def default(self) -> typing.Any:
        if isinstance(self.converter, _ConsumeRestConverter):
            # can't do the ternary here as for some reason, mypy can't tell if converter has "param_name" attr
            return {self.converter.param_name: self._default}

        return self._default

    @default.setter
    def default(self, default: typing.Any) -> None:
        self._default = default

    @staticmethod
    def apply_default_recursively(converter: typing.Optional[_BaseConverter], default: typing.Any) -> None:
        if converter is None:
            return None

        if isinstance(converter, _DefaultingConverter):
            converter._default = default

        return _DefaultingConverter.apply_default_recursively(getattr(converter, "converter", None), default)

    async def convert(
        self, context: context_.Context, arg_string: str, *, parse: bool = False
    ) -> typing.Tuple[typing.Union[dict, T], str]:
        if not arg_string:
            return self.default, ""

        try:
            converted_arg = await self.converter.convert(context, arg_string, parse=True)
        except Exception:
            if self.raise_on_fail:
                raise

            return self.default, arg_string

        return converted_arg


class _ConsumeRestConverter:
    __slots__ = ("converter", "param_name")

    def __init__(self, converter: _BaseConverter[T], param_name: str):
        self.converter = converter
        self.param_name = param_name

    async def convert(
        self, context: context_.Context, arg_string: str, *, parse: bool = False
    ) -> typing.Tuple[typing.Dict[str, T], str]:
        converted_arg = await self.converter.convert(context, arg_string, parse=False)
        return {self.param_name: converted_arg[0]}, ""


if typing.TYPE_CHECKING:
    user_converter = hikari.User
    member_converter = hikari.Member
    text_channel_converter = hikari.TextableChannel
    guild_voice_channel_converter = hikari.GuildVoiceChannel
    voice_channel_converter = guild_voice_channel_converter
    category_converter = hikari.GuildCategory
    role_converter = hikari.Role
    emoji_converter = hikari.Emoji
    guild_converter = hikari.GuildPreview
    message_converter = hikari.Message
    invite_converter = hikari.Invite
    colour_converter = hikari.Colour
    color_converter = colour_converter



# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__: typing.Final[typing.List[str]] = ["CommandErrorEvent", "CommandInvocationEvent", "CommandCompletionEvent"]

import abc
import typing

import attr
import hikari
from hikari.events import base_events as hikari_base_events

if typing.TYPE_CHECKING:
    import types

    from lightbulb import command_handler
    from lightbulb import commands
    from lightbulb import context as context_
    from lightbulb import errors


@attr.s(slots=True, weakref_slot=False)
@hikari_base_events.requires_intents(hikari.Intents.DM_MESSAGES, hikari.Intents.GUILD_MESSAGES)
class LightbulbEvent(hikari.Event, abc.ABC):
    """
    The base class for all lightbulb events. Every event dispatched by lightbulb
    will be an instance of a subclass of this.
    """

    app: command_handler.Bot = attr.ib()
    """Bot instance for this event."""

    @property
    def bot(self) -> command_handler.Bot:
        """Bot instance for this event. Alias for :obj:`~LightbulbEvent.app`."""
        return self.app


@attr.s(kw_only=True, slots=True, weakref_slot=False)
@hikari_base_events.requires_intents(hikari.Intents.DM_MESSAGES, hikari.Intents.GUILD_MESSAGES)
class CommandErrorEvent(LightbulbEvent):
    """
    Event type to subscribe to for the processing of all command errors raised by the handler.

    Example:

        .. code-block:: python

            from lightbulb.events import CommandErrorEvent

            bot = lightbulb.Bot(token="token_here", prefix="!")

            @bot.listen(CommandErrorEvent)
            async def handle_command_error(event):
                ...

    """

    exception: errors.LightbulbError = attr.ib()
    """The exception that triggered this event."""
    context: typing.Optional[context_.Context] = attr.ib(default=None)
    """The context that this event was triggered for. Will be ``None`` for :obj:`~CommandNotFound` errors."""
    message: hikari.Message = attr.ib()
    """The message that this event was triggered for."""
    command: typing.Optional[commands.Command] = attr.ib(default=None)
    """The command that this event was triggered for."""

    @property
    def exc_info(
        self,
    ) -> typing.Tuple[typing.Type[errors.LightbulbError], errors.LightbulbError, typing.Optional[types.TracebackType]]:
        """The exception triplet compatible with context managers and :mod:`traceback` helpers."""
        return type(self.exception), self.exception, self.exception.__traceback__

    @property
    def traceback(self) -> types.TracebackType:
        """The traceback for this event's exception."""
        return self.exception.__traceback__


@attr.s(kw_only=True, slots=True, weakref_slot=False)
@hikari_base_events.requires_intents(hikari.Intents.DM_MESSAGES, hikari.Intents.GUILD_MESSAGES)
class CommandInvocationEvent(LightbulbEvent):
    """
    Event dispatched when a command is invoked, regardless of whether or not the checks
    passed or failed, or an error was raised during command invocation.
    """

    command: commands.Command = attr.ib()
    """The command that this event was triggered for."""
    context: context_.Context = attr.ib()
    """The context that this event was triggered for."""


@attr.s(kw_only=True, slots=True, weakref_slot=False)
@hikari_base_events.requires_intents(hikari.Intents.DM_MESSAGES, hikari.Intents.GUILD_MESSAGES)
class CommandCompletionEvent(LightbulbEvent):
    """
    Event type dispatched when a command invocation occurred and was completed successfully. This means
    that all checks must have passed and that no errors can have been raised during the command invocation.
    """

    command: commands.Command = attr.ib()
    """The command that this event was triggered for."""
    context: context_.Context = attr.ib()
    """The context that this event was triggered for."""



# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
"""
.. error::
    Note that all check decorators **must** be above the command decorator otherwise your code will not work.
"""
from __future__ import annotations

__all__: typing.Final[typing.List[str]] = [
    "dm_only",
    "guild_only",
    "owner_only",
    "bot_only",
    "webhook_only",
    "human_only",
    "has_roles",
    "has_guild_permissions",
    "bot_has_guild_permissions",
    "has_permissions",
    "bot_has_permissions",
    "has_attachment",
    "check",
    "check_exempt",
]

import functools
import inspect
import operator
import typing

import hikari

from lightbulb import commands
from lightbulb import context
from lightbulb import errors

if typing.TYPE_CHECKING:
    from hikari import snowflakes

T_inv = typing.TypeVar("T_inv", bound=commands.Command)

_CHECK_DECORATOR_BELOW_COMMAND_DECORATOR_MESSAGE = """Check decorators MUST be above the commands decorator in order for them to work.

Valid:

@lightbulb.guild_only()
@lightbulb.command()
async def foo(ctx):
    ...

Invalid:

@lightbulb.command()
@lightbulb.guild_only()
async def foo(ctx):
    ...

See https://hikari-lightbulb.readthedocs.io/en/latest/api-reference.html#module-lightbulb.checks
"""


def _check_check_decorator_above_commands_decorator(func_or_command) -> None:
    if inspect.isfunction(func_or_command) or inspect.ismethod(func_or_command):
        raise SyntaxError(_CHECK_DECORATOR_BELOW_COMMAND_DECORATOR_MESSAGE)


async def _guild_only(ctx: context.Context) -> bool:
    if ctx.message.guild_id is None:
        raise errors.OnlyInGuild("This command can only be used in a guild")
    return True


async def _dm_only(ctx: context.Context) -> bool:
    if ctx.message.guild_id is not None:
        raise errors.OnlyInDM("This command can only be used in DMs")
    return True


async def _owner_only(ctx: context.Context) -> bool:
    if not ctx.bot.owner_ids:
        await ctx.bot.fetch_owner_ids()

    if ctx.message.author.id not in ctx.bot.owner_ids:
        raise errors.NotOwner("You are not the owner of this bot")
    return True


async def _bot_only(ctx: context.Context) -> bool:
    if not ctx.author.is_bot:
        raise errors.BotOnly(f"Only a bot can use {ctx.invoked_with}")
    return True


async def _webhook_only(ctx: context.Context) -> bool:
    if ctx.message.webhook_id is None:
        raise errors.WebhookOnly(f"Only a webhook can use {ctx.invoked_with}")
    return True


async def _human_only(ctx: context.Context) -> bool:
    if ctx.author.is_bot or ctx.message.webhook_id is not None:
        raise errors.HumanOnly(f"Only an human can use {ctx.invoked_with}")
    return True


async def _nsfw_channel_only(ctx: context.Context) -> bool:
    if not ctx.channel.is_nsfw:
        raise errors.NSFWChannelOnly(f"{ctx.invoked_with} can only be used in an NSFW channel")
    return True


def _role_check(member_roles: typing.Sequence[hikari.Snowflake], *, roles: typing.Sequence[int], func) -> bool:
    return func(r in member_roles for r in roles)


async def _has_roles(ctx: context.Context, *, role_check):
    await _guild_only(ctx)
    if not role_check(ctx.member.role_ids):
        raise errors.MissingRequiredRole("You are missing one or more roles required in order to run this command.")
    return True


def _get_missing_perms(permissions: hikari.Permissions, roles: typing.Sequence[hikari.Role]) -> hikari.Permissions:
    missing_perms = hikari.Permissions.NONE
    user_permissions = set([role.permissions for role in roles])

    if hikari.Permissions.ADMINISTRATOR in user_permissions:
        return hikari.Permissions.NONE

    for required_permission in permissions:
        if required_permission not in user_permissions:
            missing_perms |= required_permission

    return missing_perms


async def _has_guild_permissions(ctx: context.Context, *, permissions: hikari.Permissions):
    if not (ctx.bot.intents & hikari.Intents.GUILDS) == hikari.Intents.GUILDS:
        raise hikari.MissingIntentError(hikari.Intents.GUILDS)

    await _guild_only(ctx)

    # Checks if the author is the server owner
    if ctx.guild.owner_id == ctx.author.id:
        return True

    roles = ctx.bot.cache.get_roles_view_for_guild(ctx.guild_id).values()

    missing_perms = _get_missing_perms(permissions, [role for role in roles if role.id in ctx.member.role_ids])

    if missing_perms:
        raise errors.MissingRequiredPermission(
            "You are missing one or more permissions required in order to run this command", missing_perms
        )
    return True


async def _bot_has_guild_permissions(ctx: context.Context, *, permissions: hikari.Permissions):
    if not (ctx.bot.intents & hikari.Intents.GUILDS) == hikari.Intents.GUILDS:
        raise hikari.MissingIntentError(hikari.Intents.GUILDS)

    await _guild_only(ctx)

    if ctx.guild.owner_id == ctx.bot.cache.get_me().id:
        return True

    roles = ctx.bot.cache.get_roles_view_for_guild(ctx.guild_id).values()
    bot_member = ctx.bot.cache.get_member(ctx.guild_id, ctx.bot.cache.get_me().id)

    missing_perms = _get_missing_perms(permissions, [role for role in roles if role.id in bot_member.role_ids])

    if missing_perms:
        raise errors.BotMissingRequiredPermission(
            "I am missing one or more permissions required in order to run this command", missing_perms
        )
    return True


async def _has_permissions(ctx: context.Context, *, permissions: hikari.Permissions) -> bool:
    if not (ctx.bot.intents & hikari.Intents.GUILDS) == hikari.Intents.GUILDS:
        raise hikari.MissingIntentError(hikari.Intents.GUILDS)

    await _guild_only(ctx)

    perm_over = ctx.channel.permission_overwrites.values()
    perm_none = hikari.Permissions.NONE

    allowed_perms = functools.reduce(
        operator.or_, (override.allow if override.id in ctx.member.role_ids else perm_none for override in perm_over)
    )

    missing_perms = allowed_perms ^ permissions

    if missing_perms:
        raise errors.MissingRequiredPermission(
            "You are missing one or more permissions required in order to run this command", missing_perms
        )
    return True


async def _bot_has_permissions(ctx: context.Context, *, permissions: hikari.Permissions) -> bool:
    if not (ctx.bot.intents & hikari.Intents.GUILDS) == hikari.Intents.GUILDS:
        raise hikari.MissingIntentError(hikari.Intents.GUILDS)

    await _guild_only(ctx)

    perm_over = ctx.channel.permission_overwrites.values()
    perm_none = hikari.Permissions.NONE
    bot_member = ctx.bot.cache.get_member(ctx.guild_id, ctx.bot.cache.get_me().id)

    allowed_perms = functools.reduce(
        operator.or_, (override.allow if override.id in bot_member.role_ids else perm_none for override in perm_over)
    )

    missing_perms = allowed_perms ^ permissions

    if missing_perms:
        raise errors.BotMissingRequiredPermission(
            "I am missing one or more permissions required in order to run this command", missing_perms
        )
    return True


async def _has_attachment(
    ctx: context.Context, *, allowed_extensions: typing.Optional[typing.Sequence[str]] = None
) -> bool:
    if not ctx.attachments:
        raise errors.MissingRequiredAttachment("Missing attachment(s) required to run the command")

    if allowed_extensions is not None:
        for attachment in ctx.attachments:
            if not any(attachment.filename.endswith(ext) for ext in allowed_extensions):
                raise errors.MissingRequiredAttachment("Missing attachment(s) required to run the command")

    return True


def guild_only() -> typing.Callable[[T_inv], T_inv]:
    """
    A decorator that prevents a command from being used in direct messages.
    """

    def decorate(command: T_inv) -> T_inv:
        _check_check_decorator_above_commands_decorator(command)
        command.add_check(_guild_only)
        return command

    return decorate


def dm_only() -> typing.Callable[[T_inv], T_inv]:
    """
    A decorator that prevents a command from being used in a guild.

    Example:

        .. code-block:: python

            @lightbulb.dm_only()
            @bot.command()
            async def foo(ctx):
                await ctx.respond("bar")
    """

    def decorate(command: T_inv) -> T_inv:
        _check_check_decorator_above_commands_decorator(command)
        command.add_check(_dm_only)
        return command

    return decorate


def owner_only() -> typing.Callable[[T_inv], T_inv]:
    """
    A decorator that prevents a command from being used by anyone other than the owner of the application.
    """

    def decorate(command: T_inv) -> T_inv:
        _check_check_decorator_above_commands_decorator(command)
        command.add_check(_owner_only)
        return command

    return decorate


def bot_only() -> typing.Callable[[T_inv], T_inv]:
    """
    A decorator that prevents a command from being used by anyone other than a bot.
    """

    def decorate(command: T_inv) -> T_inv:
        _check_check_decorator_above_commands_decorator(command)
        command.add_check(_bot_only)
        return command

    return decorate


def webhook_only() -> typing.Callable[[T_inv], T_inv]:
    """
    A decorator that prevents a command from being used by anyone other than a webhook.
    """

    def decorate(command: T_inv) -> T_inv:
        _check_check_decorator_above_commands_decorator(command)
        command.add_check(_webhook_only)
        return command

    return decorate


def human_only() -> typing.Callable[[T_inv], T_inv]:
    """
    A decorator that prevents a command from being used by anyone other than a human.
    """

    def decorate(command: T_inv) -> T_inv:
        _check_check_decorator_above_commands_decorator(command)
        command.add_check(_human_only)
        return command

    return decorate


def nsfw_channel_only() -> typing.Callable[[T_inv], T_inv]:
    def decorate(command: T_inv) -> T_inv:
        _check_check_decorator_above_commands_decorator(command)
        command.add_check(_nsfw_channel_only)
        return command

    return decorate


def has_roles(
    role1: snowflakes.SnowflakeishOr[hikari.PartialRole],
    *role_ids: snowflakes.SnowflakeishOr[hikari.PartialRole],
    mode: typing.Literal["all", "any"] = "all",
):
    """
    A decorator that prevents a command from being used by anyone missing roles according
    to the given mode.

    Args:
        role1 (:obj:`~hikari.snowflakes.SnowflakeishOr` [ :obj:`~hikari.PartialRole` ]): Role ID to check for.
        *role_ids (:obj:`~hikari.snowflakes.SnowflakeishOr` [ :obj:`~hikari.PartialRole` ]): Additional role IDs to check for.

    Keyword Args:
        mode (Literal["all", "any"]): The mode to check roles using. If ``"all"``, all role IDs
            passed will be required. If ``"any"``, the invoker will only be required to have one
            of the specified roles. Defaults to ``"all"``.

    Note:
        This check will also prevent commands from being used in DMs, as you cannot have roles
        in a DM channel.
    """
    if mode not in ["all", "any"]:
        raise SyntaxError("has_roles mode must be one of: all, any")

    def decorate(command: T_inv) -> T_inv:
        _check_check_decorator_above_commands_decorator(command)
        check_func = functools.partial(
            _role_check, roles=[int(role1), *[int(r) for r in role_ids]], func=all if mode == "all" else any
        )
        command.add_check(functools.partial(_has_roles, role_check=check_func))
        return command

    return decorate


def has_guild_permissions(perm1: hikari.Permissions, *permissions: hikari.Permissions):
    """
    A decorator that prevents the command from being used by a member missing any of the required
    guild permissions (permissions granted by a role).

    Args:
        perm1 (:obj:`hikari.Permissions`): Permission to check for.
        *permissions (:obj:`hikari.Permissions`): Additional permissions to check for.

    Note:
        This check will also prevent commands from being used in DMs, as you cannot have permissions
        in a DM channel.

    Warning:
        This check is unavailable if your application is stateless and/or missing the intent
        :obj:`hikari.Intents.GUILDS` and will **always** raise an error on command invocation if
        either of these conditions are not met.
    """

    def decorate(command: T_inv) -> T_inv:
        _check_check_decorator_above_commands_decorator(command)
        perms = perm1.split()

        total_perms = functools.reduce(operator.or_, (*perms, *permissions))
        command.user_required_permissions = total_perms

        command.add_check(functools.partial(_has_guild_permissions, permissions=total_perms))
        return command

    return decorate


def bot_has_guild_permissions(perm1: hikari.Permissions, *permissions: hikari.Permissions):
    """
    A decorator that prevents the command from being used if the bot is missing any of the required
    guild permissions (permissions granted by a role).

    Args:
        perm1 (:obj:`hikari.Permissions`): Permission to check for.
        *permissions (:obj:`hikari.Permissions`): Additional permissions to check for.

    Note:
        This check will also prevent commands from being used in DMs, as you cannot have permissions
        in a DM channel.

    Warning:
        This check is unavailable if your application is stateless and/or missing the intent
        :obj:`hikari.Intents.GUILDS` and will **always** raise an error on command invocation if
        either of these conditions are not met.
    """

    def decorate(command: T_inv) -> T_inv:
        _check_check_decorator_above_commands_decorator(command)
        perms = perm1.split()

        total_perms = functools.reduce(operator.or_, (*perms, *permissions))
        command.bot_required_permissions = total_perms

        command.add_check(functools.partial(_bot_has_guild_permissions, permissions=total_perms))
        return command

    return decorate


def has_permissions(perm1: hikari.Permissions, *permissions: hikari.Permissions):
    """
    A decorator that prevents the command from being used by a member missing any of the required
    channel permissions (permissions granted by a permission overwrite).

    Args:
        perm1 (:obj:`hikari.Permissions`): Permission to check for.
        *permissions (:obj:`hikari.Permissions`): Additional permissions to check for.

    Note:
        This check will also prevent commands from being used in DMs, as you cannot have permissions
        in a DM channel.
    """

    def decorate(command: T_inv) -> T_inv:
        _check_check_decorator_above_commands_decorator(command)
        perms = perm1.split()

        total_perms = functools.reduce(operator.or_, (*perms, *permissions))
        command.user_required_permissions = total_perms

        command.add_check(functools.partial(_has_permissions, permissions=total_perms))
        return command

    return decorate


def bot_has_permissions(perm1: hikari.Permissions, *permissions: hikari.Permissions):
    """
    A decorator that prevents the command from being used if the bot is missing any of the required
    channel permissions (permissions granted by a permission overwrite).

    Args:
        perm1 (:obj:`hikari.Permissions`): Permission to check for.
        *permissions (:obj:`hikari.Permissions`): Additional permissions to check for.

    Note:
        This check will also prevent commands from being used in DMs, as you cannot have permissions
        in a DM channel.
    """

    def decorate(command: T_inv) -> T_inv:
        _check_check_decorator_above_commands_decorator(command)
        perms = perm1.split()

        total_perms = functools.reduce(operator.or_, (*perms, *permissions))
        command.user_required_permissions = total_perms

        command.add_check(functools.partial(_bot_has_permissions, permissions=total_perms))
        return command

    return decorate


def has_attachment(*extensions: str):
    """
    A decorator that prevents the command from being used if the invocation message
    does not include any attachments.

    Args:
        *extensions (:obj:`str`): If specified, attachments with a different file extension
            will cause the check to fail.

    Example:

        .. code-block:: python

            @checks.has_attachment(".yaml", ".json")
            @bot.command()
            async def foobar(ctx):
                print(ctx.attachments[0].filename)

    Note:
        If ``extensions`` is specified then all attachments must conform to the restriction.
    """

    def decorate(command: T_inv) -> T_inv:
        _check_check_decorator_above_commands_decorator(command)
        command.add_check(functools.partial(_has_attachment, allowed_extensions=extensions if extensions else None))
        return command

    return decorate


def check(check_func: typing.Callable[[context.Context], typing.Coroutine[typing.Any, typing.Any, bool]]):
    """
    A decorator which adds a custom check function to a command. The check function must be a coroutine (async def)
    and take a single argument, which will be the command context.

    This acts as a shortcut to calling :meth:`~.commands.Command.add_check` on a command instance.

    Args:
        check_func (Callable[ [ :obj:`~.context.Context` ], Coroutine[ Any, Any, :obj:`bool` ] ]): The coroutine
            to add to the command as a check.

    Example:

        .. code-block:: python

            async def check_message_contains_hello(ctx):
                return "hello" in ctx.message.content

            @checks.check(check_message_contains_hello)
            @bot.command()
            async def foo(ctx):
                await ctx.respond("Bar")

    See Also:
        :meth:`~.commands.Command.add_check`
    """

    def decorate(command: T_inv) -> T_inv:
        _check_check_decorator_above_commands_decorator(command)
        command.add_check(check_func)
        return command

    return decorate


def check_exempt(predicate):
    """
    A decorator which allows **all** checks to be bypassed if the ``predicate`` conditions are met.
    Predicate can be a coroutine or a normal function but must take a single
    argument - ``context`` - and must return a boolean - ``True`` if checks should be bypassed or ``False`` if not.

    Args:
        predicate: The callable which determines if command checks should be bypassed or not.

    Example:

        .. code-block:: python

            @checks.check_exempt(lambda ctx: ctx.author.id == 1234)
            @checks.guild_only()
            @bot.command()
            async def foo(ctx):
                await ctx.respond("foo")
    """

    def decorate(command: T_inv) -> T_inv:
        _check_check_decorator_above_commands_decorator(command)
        command._check_exempt_predicate = predicate
        return command

    return decorate



# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__: typing.Final[typing.List[str]] = [
    "SignatureInspector",
    "Command",
    "Group",
    "command",
    "group",
]

import functools
import inspect
import logging
import typing

import hikari
from multidict import CIMultiDict

from lightbulb import context as context_
from lightbulb import converters
from lightbulb import cooldowns
from lightbulb import errors
from lightbulb import events
from lightbulb.converters import _BaseConverter
from lightbulb.converters import _ConsumeRestConverter
from lightbulb.converters import _Converter
from lightbulb.converters import _DefaultingConverter
from lightbulb.converters import _GreedyConverter
from lightbulb.converters import _UnionConverter

if typing.TYPE_CHECKING:
    from lightbulb import plugins

_LOGGER = logging.getLogger("lightbulb")

_CommandT = typing.TypeVar("_CommandT", bound="Command")
T = typing.TypeVar("T")
_ConverterT = typing.Union[_BaseConverter[T], _BaseConverter[typing.List[T]]]

_CONVERTER_CLASS_MAPPING = {
    hikari.User: converters.user_converter,
    hikari.Member: converters.member_converter,
    hikari.TextableChannel: converters.text_channel_converter,
    hikari.GuildVoiceChannel: converters.guild_voice_channel_converter,
    hikari.GuildCategory: converters.category_converter,
    hikari.Role: converters.role_converter,
    hikari.Emoji: converters.emoji_converter,
    hikari.GuildPreview: converters.guild_converter,
    hikari.Message: converters.message_converter,
    hikari.Invite: converters.invite_converter,
    hikari.Colour: converters.colour_converter,
    hikari.Color: converters.colour_converter,
}


class _BoundCommandMarker:
    def __init__(self, delegates_to: _CommandT) -> None:
        self.delegates_to = delegates_to


def _bind_prototype(instance: typing.Any, command_template: _CommandT):
    # Create a prototype of the command which is bound to the given instance.

    class BoundCommand(type(command_template), _BoundCommandMarker):
        def __init__(self) -> None:
            _BoundCommandMarker.__init__(self, command_template)
            # Do not init the super class, simply delegate to it using the __dict__
            self._delegates_to = command_template
            self.__dict__.update(command_template.__dict__)

        def __hash__(self) -> int:
            return hash(self.name)

        def __eq__(self, other) -> bool:
            return isinstance(other, (type(self), Command)) and other.name == self.name

        async def invoke(self, context: context_.Context, *args: str, **kwargs: str) -> typing.Any:
            if self.cooldown_manager is not None:
                await self.cooldown_manager.add_cooldown(context)
            # Add the start slice on to the length to offset the section of arg_details being extracted

            return await self._callback(instance, context, *args, **kwargs)

    prototype = BoundCommand()

    # This will cache this for a later call!
    instance.__dict__[command_template.method_name] = prototype

    # Bind each subcommand to a descriptor for this specific instance.
    if isinstance(prototype, Group):
        prototype._subcommands = {}
        for subcommand_name, subcommand in command_template._subcommands.items():
            for name, member in inspect.getmembers(instance, lambda m: isinstance(m, _BoundCommandMarker)):
                if member.delegates_to is subcommand:
                    # This will bind the instance to a bound method, and replace the parent. This completes the
                    # prototype, detatching it entirely from the class-bound implementation it was created from. This
                    # means adding the same plugin twice would attempt to add two unique copies of the command that
                    # hopefully are not aware of eachother by design, reducing weird side effects from shared attributes
                    # hopefully!
                    member.parent = prototype
                    prototype._subcommands[subcommand_name] = member
                    prototype.subcommands.add(member)

    return typing.cast(_CommandT, prototype)


class SignatureInspector:
    """
    Contains information about the arguments that a command takes when
    it is invoked.

    Args:
        command (:obj:`~.commands.Command`): The command to inspect the arguments of.
    """

    def __init__(self, command: Command) -> None:
        self.command = command
        self.has_self = isinstance(command, _BoundCommandMarker)
        signature = inspect.signature(command._callback)
        self.converters = self.parse_signature(signature)
        self.arguments = [p.name for p in signature.parameters.values()]
        self.arguments.pop(0)
        if self.has_self:
            self.arguments.pop(0)

    def get_converter(self, annotation) -> _ConverterT:
        """
        Resolve the converter order for a given type annotation recursively.

        Args:
            annotation: The parameter's type annotation.

        Returns:
            The top level converter for the parameter
        """
        annotation = _CONVERTER_CLASS_MAPPING.get(annotation, annotation)
        if isinstance(annotation, str):
            annotation = str

        origin = typing.get_origin(annotation)
        args = typing.get_args(annotation)

        if origin is typing.Union:
            convs: typing.List[_ConverterT] = []

            for arg in args:
                if arg is None.__class__:
                    return _DefaultingConverter(_UnionConverter(*convs), None, raise_on_fail=False)

                convs.append(self.get_converter(arg))

            return _UnionConverter(*convs)

        if origin is converters.Greedy:
            return _GreedyConverter(self.get_converter(args[0]))

        return _Converter(annotation if annotation is not inspect.Parameter.empty else str)

    def parse_signature(self, signature: inspect.Signature):
        """
        Parse the command's callback signature into a list of the converters used
        for each command argument.

        Args:
            signature (:obj:`inspect.Signature`): The signature of the command callback.

        Returns:
            List of converters for command arguments in the correct order.
        """
        arg_converters = []

        for idx, param in enumerate(signature.parameters.values()):
            if idx == 0 or (idx == 1 and self.has_self):
                continue

            converter: _BaseConverter = self.get_converter(param.annotation)

            if param.kind is inspect.Parameter.KEYWORD_ONLY:
                converter = _ConsumeRestConverter(converter, param.name)
            elif param.kind is inspect.Parameter.VAR_POSITIONAL:
                converter = _GreedyConverter(converter, unpack=True)

            if param.default is not inspect.Parameter.empty:
                if not isinstance(converter, _DefaultingConverter):
                    converter = _DefaultingConverter(converter, param.default)
                else:
                    converter.default = param.default

            arg_converters.append(converter)
        return arg_converters


class Command:
    """
    A command that can be invoked by a user. When invoked, the callback
    will be called with a set argument ctx, an instance of the :obj:`~.context.Context`
    class, and any other arguments supplied by the user.

    Args:
        callback: The coroutine to register as the command's callback.
        name (:obj:`str`): The name to register the command to.
        allow_extra_arguments (:obj:`bool`): Whether or not the command should error
            when run with too many arguments.
        aliases (Iterable[ :obj:`str` ]): Additional names to register the command to.
        hidden (:obj:`bool`): Whether or not to show the command in the bot's help overview.
        parent (Optional[ :obj:`~Group` ]): The parent group for the command, or ``None`` if
            the command has no parent.
    """

    def __init__(
        self,
        callback: typing.Callable,
        name: str,
        allow_extra_arguments: bool,
        aliases: typing.Iterable[str],
        hidden: bool,
        parent: typing.Optional[typing.Any] = None,
    ) -> None:
        self._callback = callback
        self._name = name
        self._allow_extra_arguments = allow_extra_arguments
        self._aliases = aliases
        self.hidden = hidden
        self._checks = []
        self._check_exempt_predicate = lambda ctx: False
        self._raw_error_listener = None
        self._raw_before_invoke = None
        self._raw_after_invoke = None
        self.method_name: typing.Optional[str] = None
        self.parent: typing.Optional[Group] = parent
        """The parent group for the command. If ``None`` then the command is not a subcommand."""
        self.plugin: typing.Optional[plugins.Plugin] = None
        """The plugin the command is registered to. If ``None`` then it was defined outside of a plugin."""
        self.user_required_permissions: hikari.Permissions = hikari.Permissions.NONE
        """
        The permissions required by a user to run the command.
        These are extracted from the permission check decorator(s) on the command.
        """
        self.bot_required_permissions: hikari.Permissions = hikari.Permissions.NONE
        """
        The permissions the bot requires for a user to be able to run the command.
        These are extracted from the permission check decorator(s) on the command.
        """

        self.cooldown_manager: typing.Optional[cooldowns.CooldownManager] = None
        """The cooldown manager being used for the command. If ``None`` then the command does not have a cooldown."""

        signature = inspect.signature(callback)
        self._has_max_args = not any(
            a.kind == inspect.Parameter.VAR_POSITIONAL
            or a.kind == inspect.Parameter.KEYWORD_ONLY
            and a.default == inspect.Parameter.empty
            for a in signature.parameters.values()
        )

    def __get__(self: _CommandT, instance: typing.Any, owner: typing.Type[typing.Any]) -> _CommandT:
        return _bind_prototype(instance, self)

    def __set_name__(self, owner, name):
        self.method_name = name

    def __repr__(self) -> str:
        return f"<lightbulb.{self.__class__.__name__} {self.name} at {hex(id(self))}>"

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other) -> bool:
        return isinstance(other, type(self)) and other.name == self.name

    @functools.cached_property
    def _error_listener(self):
        if self.plugin is not None and self._raw_error_listener is not None:
            return getattr(self.plugin, self._raw_error_listener.__name__)
        return self._raw_error_listener

    @functools.cached_property
    def _before_invoke(self):
        if self.plugin is not None and self._raw_before_invoke is not None:
            return getattr(self.plugin, self._raw_before_invoke.__name__)
        return self._raw_before_invoke

    @functools.cached_property
    def _after_invoke(self):
        if self.plugin is not None and self._raw_after_invoke is not None:
            return getattr(self.plugin, self._raw_after_invoke.__name__)
        return self._raw_after_invoke

    @functools.cached_property
    def arg_details(self) -> SignatureInspector:
        """
        An inspection of the arguments that a command takes.

        Returns:
            :obj:`~.commands.SignatureInspector`: Details about the command's arguments.
        """
        return SignatureInspector(self)

    @property
    def name(self) -> str:
        """
        The registered name of the command

        Returns:
            :obj:`str`: Name of the command
        """
        return self._name

    @property
    def qualified_name(self) -> str:
        """
        The fully qualified name of the command, taking into account
        whether or not the command is a subcommand.

        Returns:
            :obj:`str`: Qualified name of the command.
        """
        command_qualname = []
        cmd = self
        while cmd is not None:
            command_qualname.append(cmd.name)
            cmd = cmd.parent
        return " ".join(command_qualname[::-1])

    @property
    def aliases(self) -> typing.Iterable[str]:
        """
        The command's aliases.

        Returns:
            Iterable[ :obj:`str` ]: Aliases for the command.
        """
        return self._aliases

    @property
    def is_subcommand(self) -> bool:
        """
        Whether or not the object is a subcommand with a parent group

        Returns:
            Union[ :obj:`True`, :obj:`False` ]: If this object is a subcommand with a parent group.
        """
        return self.parent is not None

    @property
    def callback(self) -> typing.Callable:
        """
        The command's callback coroutine

        Returns:
            Callable: The callback coroutine for the command
        """
        return self._callback

    @property
    def checks(self) -> typing.Iterable[typing.Callable[[context_.Context], bool]]:
        """
        The command's checks.

        Returns:
            Iterable[ Callable[ [ :obj:`~.context.Context` ], :obj:`bool` ] ]: The checks for the command.
        """
        return self._checks

    def command_error(self):
        """
        A decorator to register a coroutine as the command's error handler. Any return from the error handler
        will be used to determine whether or not the error was handled successfully. A return of any truthy
        object will prevent the event from propagating further down the listener chain. Returning a falsy value
        will mean that any global command error listener will then be called.

        The coroutine can only take one argument, which will be an instance of :obj:`~.events.CommandErrorEvent`
        unless in a class in which case it may also take the ``self`` argument.

        Example:

            .. code-block:: python

                @lightbulb.owner_only()
                @bot.command()
                async def foo(ctx):
                    await ctx.respond("bar")

                @foo.command_error()
                async def on_foo_error(event):
                    # This will be called if someone other than the
                    # bot's owner tries to run the command foo
                    return
        """

        def decorate(
            func: typing.Callable[[events.CommandErrorEvent], typing.Coroutine[None, None, typing.Any]]
        ) -> typing.Callable[[events.CommandErrorEvent], typing.Coroutine[None, None, typing.Any]]:
            self._raw_error_listener = func
            return func

        return decorate

    def before_invoke(self):
        """
        A decorator to register a coroutine to be run before the command is invoked.

        The coroutine can take only one argument which will be an instance of the :obj:`~.context.Context`
        for the command.
        """

        def decorate(
            func: typing.Callable[[context_.Context], typing.Coroutine[None, None, typing.Any]]
        ) -> typing.Callable[[context_.Context], typing.Coroutine[None, None, typing.Any]]:
            self._raw_before_invoke = func
            return func

        return decorate

    def after_invoke(self):
        """
        A decorator to register a coroutine to be run after the command is invoked.

        The coroutine can take only one argument which will be an instance of the :obj:`~.context.Context`
        for the command.
        """

        def decorate(
            func: typing.Callable[[context_.Context], typing.Coroutine[None, None, typing.Any]]
        ) -> typing.Callable[[context_.Context], typing.Coroutine[None, None, typing.Any]]:
            self._raw_after_invoke = func
            return func

        return decorate

    async def invoke(self, context: context_.Context, *args: str, **kwargs: str) -> typing.Any:
        """
        Invoke the command with given args and kwargs. Cooldowns and converters will
        be processed however this method bypasses all command checks.

        Args:
            context (:obj:`~.context.Context`): The command invocation context.
            *args: The positional arguments to invoke the command with.

        Keyword Args:
            **kwargs: The keyword arguments to invoke the command with.

        """
        if self.cooldown_manager is not None:
            await self.cooldown_manager.add_cooldown(context)

        return await self._callback(context, *args, **kwargs)

    def add_check(self, check_func: typing.Callable[[context_.Context], typing.Coroutine[None, None, bool]]) -> None:
        """
        Add a check to an instance of :obj:`~.commands.Command` or a subclass. The check passed must
        be an awaitable function taking a single argument which will be an instance of :obj:`~.context.Context`.
        It must also either return a boolean denoting whether or not the check passed,
        or raise an instance of :obj:`~.errors.CheckFailure` or a subclass.

        Args:
            check_func (Callable[ [ :obj:`~.context.Context` ], Coroutine[ ``None``, ``None``, :obj:`bool` ] ]): Check
                to add to the command

        Returns:
            ``None``

        Example:

            .. code-block:: python

                async def author_name_startswith_foo(ctx):
                    return ctx.author.username.startswith("foo")

                bot.get_command("foo").add_check(author_name_startswith_foo)

        See Also:
            :meth:`~.checks.check`
        """
        self._checks.append(check_func)

    async def is_runnable(self, context: context_.Context) -> bool:
        """
        Run all the checks for the command to determine whether or not it is
        runnable in the given context.

        Args:
            context (:obj:`~.context.Context`): The context to evaluate the checks with.

        Returns:
            :obj:`bool`: If the command is runnable in the context given.

        Raises:
            :obj:`~.errors.CheckFailure`: If the command is not runnable in the context given.
        """
        for check in self._checks:
            try:
                result = await check(context)
            except errors.MissingRequiredAttachment:
                # Bypass attachment check because we still want the command to show for the help
                # overview and it is unlikely that someone will call the help command with an attachment
                # to the message.
                result = True
            if result is False:
                raise errors.CheckFailure(f"Check {check.__name__} failed for command {self.name}")
        return True


class Group(Command):
    """
    A command group. This is invoked the same way as a normal command, however it has support
    for subcommands which can be registered to the group and invoked as separate commands.

    Args:
        *args: The args passed to :obj:`~.commands.Command` in its constructor

    Keyword Args:
        insensitive_commands (:obj:`bool`): Whether or not this group's subcommands should be case insensitive.
            Defaults to ``False``.
        inherit_checks (:obj:`bool`): Whether or not this group's subcommands should inherit the checks of the parent.
            Defaults to ``True``.
        **kwargs: The kwargs passed to :obj:`~.commands.Command` in its constructor
    """

    def __init__(self, *args, insensitive_commands: bool = False, inherit_checks: bool = True, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.insensitive_commands = insensitive_commands
        self.inherit_checks = inherit_checks
        self._subcommands: typing.MutableMapping[str, Command] = {} if not self.insensitive_commands else CIMultiDict()
        self.subcommands: typing.Set[Command] = set()
        """A set containing all subcommands registered to the group."""

    def _resolve_subcommand(self, args) -> typing.Tuple[typing.Union[Command, Group], typing.Iterable[str]]:
        this = self

        args.pop(0)
        while isinstance(this, Group) and args:
            invoked_with = args[0].casefold() if this.insensitive_commands else args[0]
            try:
                this = this._subcommands[invoked_with]
            except KeyError:
                break
            else:
                args = args[1:]

        return this, args

    def walk_commands(self) -> typing.Generator[Command, None, None]:
        """
        A generator that walks through all commands and subcommands registered to this group.

        Yields:
            :obj:`~.commands.Command`: All commands, groups and subcommands registered to this group.
        """
        for command in self.subcommands:
            yield command
            if isinstance(command, Group):
                yield from command.walk_commands()

    def add_check(self, check_func: typing.Callable[[context_.Context], typing.Coroutine[None, None, bool]]) -> None:
        if self.inherit_checks:
            for c in self.subcommands:
                c.add_check(check_func)
        super().add_check(check_func)

    def get_subcommand(self, name: str) -> Command:
        """
        Get a command object for a subcommand of the group from it's registered name.

        Args:
            name (:obj:`str`): The name of the command to get the object for.

        Returns:
            Optional[ :obj:`~.commands.Command` ]: command object registered to that name.
        """
        return self._subcommands.get(name)

    def command(self, **kwargs):
        """
        A decorator that registers a callable as a subcommand for the command group.

        Keyword Args:
            **kwargs: Kwargs passed to :obj:`~command`

        Example:

            .. code-block:: python

                bot = lightbulb.Bot(token="token_here", prefix="!")

                @bot.group()
                async def foo(ctx):
                    await ctx.respond("Invoked foo")

                @foo.command()
                async def bar(ctx):
                    await ctx.respond("Invoked foo bar")
        """

        def decorate(func):
            name = kwargs.get("name", func.__name__)
            cls = kwargs.get("cls", Command)
            self._subcommands[name] = cls(
                func,
                name,
                kwargs.get("allow_extra_arguments", True),
                kwargs.get("aliases", []),
                kwargs.get("hidden", False),
                parent=self,
            )
            if self.inherit_checks:
                self._subcommands[name]._checks.extend(self._checks)
            self.subcommands.add(self._subcommands[name])
            for alias in kwargs.get("aliases", []):
                self._subcommands[alias] = self._subcommands[name]
            return self._subcommands[name]

        return decorate

    def group(self, **kwargs):
        """
        A decorator that registers a callable as a subgroup for the command group.

        Keyword Args:
            **kwargs: Kwargs passed to :obj:`~group`
        """
        kwargs["cls"] = type(self)

        def decorate(func):
            name = kwargs.get("name", func.__name__)
            cls = kwargs.get("cls", Group)
            self._subcommands[name] = cls(
                func,
                name,
                kwargs.get("allow_extra_arguments", True),
                kwargs.get("aliases", []),
                kwargs.get("hidden", False),
                insensitive_commands=kwargs.get("insensitive_commands", False),
                inherit_checks=kwargs.get("inherit_checks", True),
                parent=self,
            )
            if self.inherit_checks:
                self._subcommands[name]._checks.extend(self._checks)
            self.subcommands.add(self._subcommands[name])
            for alias in kwargs.get("aliases", []):
                self._subcommands[alias] = self._subcommands[name]
            return self._subcommands[name]

        return decorate


def command(**kwargs):
    """
    A decorator to convert a coroutine into a :obj:`~.commands.Command` object.

    Keyword Args:
        name (Optional[ :obj:`str` ]): Name to register the command to. Defaults to the name of the coroutine.
        allow_extra_arguments (Optional[ :obj:`bool` ]): Whether or not the command should error when run with
            more arguments than it takes. Defaults to True - will not raise an error.
        aliases (Iterable[ :obj:`str` ]): Iterable of aliases which will also invoke the command.
        hidden (:obj:`bool`): Whether or not the command should be hidden from the help command. Defaults to ``False``.
        cls (:obj:`~.commands.Command`): The class to use to instantiate the command object from. Defaults
            to :obj:`~.commands.Command`.
    """

    def decorate(func):
        name = kwargs.get("name", func.__name__)
        cls = kwargs.get("cls", Command)
        return cls(
            func,
            name,
            kwargs.get("allow_extra_arguments", True),
            kwargs.get("aliases", []),
            kwargs.get("hidden", False),
        )

    return decorate


def group(**kwargs):
    """
    A decorator to convert a coroutine into a :obj:`~.commands.Group` object.

    Keyword Args:
        name (Optional[ :obj:`str` ]): Name to register the command to. Defaults to the name of the coroutine.
        allow_extra_arguments (Optional[ :obj:`bool` ]): Whether or not the command should error when run with
            more arguments than it takes. Defaults to True - will not raise an error.
        aliases (Optional[ Iterable[ :obj:`str` ] ]): Iterable of aliases which will also invoke the command.
        hidden (:obj:`bool`): Whether or not the command should be hidden from the help command. Defaults to ``False``.
        insensitive_commands (:obj:`bool`): Whether or not subcommands should be case-insensitive. Defaults to
            False.
        cls (:obj:`~.commands.Command`): The class to use to instantiate the group object from. Defaults
            to :obj:`~.commands.Group`.
        inherit_checks (:obj:`bool`): Whether or not subcommands should inherit checks added to the base group.
    """

    def decorate(func):
        name = kwargs.get("name", func.__name__)
        cls = kwargs.get("cls", Group)
        return cls(
            func,
            name,
            kwargs.get("allow_extra_arguments", True),
            kwargs.get("aliases", []),
            kwargs.get("hidden", False),
            insensitive_commands=kwargs.get("insensitive_commands", False),
            inherit_checks=kwargs.get("inherit_checks", True),
        )

    return decorate



# -*- coding: utf-8 -*-
# Copyright © Thomm.o 2021
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__: typing.Final[typing.List[str]] = ["SlashCommandOptionsWrapper", "SlashCommandContext"]

import functools
import typing

import hikari

if typing.TYPE_CHECKING:
    from lightbulb import command_handler
    from lightbulb.slash_commands import commands


class SlashCommandOptionsWrapper:
    """
    A wrapper class for :obj:`~lightbulb.slash_commands.SlashCommandContext.options` which allows the user
    to access the option values through a more user_friendly, attribute syntax.
    :obj:`~lightbulb.slash_commands.SlashCommandOptionsWrapper.option_name` will return either the option's value,
    or ``None`` if the option was not provided in the slash command invocation.

    This is accessible through :obj:`~lightbulb.slash_commands.SlashCommandContext.option_values`

    Args:
        options (Mapping[:obj:`str`, :obj:`hikari.CommandInteractionOption`): The options that the slash command
            was called with.
    """

    __slots__: typing.Sequence[str] = ("_options",)

    def __init__(self, options: typing.Mapping[str, hikari.CommandInteractionOption]) -> None:
        self._options = options

    def __getattr__(self, item: str) -> typing.Optional[typing.Union[str, int, bool, float]]:
        option = self._options.get(item)
        return option.value if option is not None else None


class SlashCommandContext:
    """
    The context a slash command was invoked under.

    Args:
        bot (:obj:`~lightbulb.command_handler.Bot`): The bot instance that received the slash command.
        interaction (:obj:`hikari.CommandInteraction`): The interaction for this slash command invocation.
        command (:obj:`~lightbulb.slash_commands.SlashCommand`): The :obj:`~SlashCommand` object that was invoked.
    """

    def __init__(
        self,
        bot: command_handler.Bot,
        interaction: hikari.CommandInteraction,
        command: commands.BaseSlashCommand,
    ) -> None:
        self.bot = bot
        """The bot instance that received the slash command."""
        self._interaction = interaction
        self._command = command

        self.options: typing.Mapping[str, hikari.CommandInteractionOption] = (
            {option.name: option for option in self._interaction.options}
            if self._interaction.options is not None
            else {}
        )
        """A mapping of :obj:`str` option name to :obj:`hikari.CommandInteractionOption` containing the options the command was invoked with."""

    @property
    def interaction(self) -> hikari.CommandInteraction:
        """The interaction for this slash command invocation."""
        return self._interaction

    @property
    def channel_id(self) -> hikari.Snowflake:
        """The channel ID that the slash command was invoked in."""
        return self._interaction.channel_id

    @property
    def guild_id(self) -> typing.Optional[hikari.Snowflake]:
        """The guild ID that the slash command was invoked in, or ``None`` if in DMs."""
        return self._interaction.guild_id

    @property
    def member(self) -> typing.Optional[hikari.Member]:
        """The :obj:`hikari.Member` object for the user that invoked the slash command, or ``None`` if in DMs."""
        return self._interaction.member

    @property
    def user(self) -> hikari.User:
        """The user object for the user that invoked the slash command."""
        return self._interaction.user

    @property
    def author(self) -> hikari.User:
        """Alias for :obj:`SlashCommandContext.user`."""
        return self.user

    @property
    def command_name(self) -> str:
        """The name of the slash command being invoked."""
        return self._interaction.command_name

    @property
    def command_id(self):
        """The ID of the slash command being invoked."""
        return self._interaction.command_id

    @property
    def command(self) -> hikari.Command:
        """The :obj:`hikari.Command` object for this specific context."""
        return self._command.get_command(self.guild_id)

    @property
    def channel(self) -> typing.Optional[hikari.GuildChannel]:
        """The cached channel that the command was invoked in, or ``None`` if not found."""
        return self._interaction.get_channel()

    @property
    def guild(self) -> typing.Optional[hikari.GatewayGuild]:
        """The cached guild that the command was invoked in, or ``None`` if not found."""
        return self.bot.cache.get_guild(self.guild_id)

    @property
    def resolved(self) -> typing.Optional[hikari.ResolvedOptionData]:
        """Mappings of the objects resolved for the provided command options."""
        return self._interaction.resolved

    @functools.cached_property
    def option_values(self) -> SlashCommandOptionsWrapper:
        """The values for the slash command's various options."""
        return SlashCommandOptionsWrapper(self.options)

    async def respond(self, content: hikari.UndefinedType = hikari.UNDEFINED, **kwargs: hikari.UndefinedType) -> None:
        """
        Alias for :obj:`hikari.CommandInteraction.create_initial_response` but without having to pass
        in the ``response_type`` (it is set to :obj:`hikari.ResponseType.MESSAGE_CREATE`) See Hikari documentation
        for kwargs you can pass in.

        Args:
            content (:obj:`hikari.UndefinedType`): The message content, generally :obj:`str`.

        Returns:
            ``None``

        Note:
            This can only be called **once** for each interaction. To add more information to the response
            you should use :obj:`~lightbulb.slash_commands.SlashCommandContext.edit_response`
        """
        await self._interaction.create_initial_response(hikari.ResponseType.MESSAGE_CREATE, content, **kwargs)

    async def edit_response(self, *args, **kwargs) -> None:
        """
        Alias for :obj:`hikari.CommandInteraction.edit_initial_response`. See Hikari documentation
        for args and kwargs you can pass in.

        Returns:
            ``None``
        """
        await self._interaction.edit_initial_response(*args, **kwargs)

    async def delete_response(self) -> None:
        """
        Alias for :obj:`hikari.CommandInteraction.delete_initial_response`.

        Returns:
            ``None``
        """
        await self._interaction.delete_initial_response()
